#!/usr/bin/env python
import logging
import os
import os.path as op
import re
import subprocess
import time
import json

from collections import deque, OrderedDict
from datetime import date
import inotify.adapters
from inotify.constants import IN_MODIFY, IN_CREATE, IN_ISDIR
from py.path import local as localpath
from tinydb import TinyDB

_DEFAULT_LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
_LOGGER = logging.getLogger(__name__)

MASK = (IN_MODIFY | IN_CREATE)
MASK_NEWDIR = (IN_CREATE | IN_ISDIR)
WAIT_TIME = 86400  # in seconds


#def _configure_logging():
_LOGGER.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
formatter = logging.Formatter(_DEFAULT_LOG_FORMAT)
ch.setFormatter(formatter)
_LOGGER.addHandler(ch)


def run_heudiconv(cmd):
    info_dict = dict()
    proc = subprocess.Popen(cmd.split(),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    return_code = proc.wait()
    if return_code == 0:
        _LOGGER.info("Done running {0}".format(cmd))
        info_dict['success'] = 1
    else:
        _LOGGER.error("{0} failed".format(cmd))
        info_dict['success'] = 0
    # get info on what we run
    stdout = proc.communicate()[0].decode('utf-8')
    match = re.match('INFO: PROCESSING STARTS: (.*)', stdout)
    info_dict_ = json.loads(match.group(1) if match else '{}')
    info_dict.update(info_dict_)
    return stdout, info_dict


def process(paths2process, db, wait=WAIT_TIME, logdir='log'):
    cmd = 'ls -l {0}'
    # if paths2process and
    # time.time() - os.path.getmtime(paths2process[0]) > WAIT_TIME:
    processed = []
    for path, mod_time in paths2process.items():
        if time.time() - mod_time > wait:
            #process_me = paths2process.popleft().decode('utf-8')
            process_me = path
            cmd_ = cmd.format(process_me)
            process_dict = {'input_path': process_me,
                            'accession_number': op.basename(process_me)}
            print("Time to process {0}".format(process_me))
            stdout, run_dict = run_heudiconv(cmd_)
            process_dict.update(run_dict)
            db.insert(process_dict)
            # save log
            logdir = localpath(logdir)
            log = logdir.join(process_dict['accession_number'] + '.log')
            log.write(stdout)
            # if we processed it, or it failed,
            # we need to remove it to avoid running it again
            processed.append(path)
    for processed_path in processed:
        del paths2process[processed_path]


def monitor(topdir='/tmp/new_dir', check_ptrn='/20../../..',
            db=None, wait=WAIT_TIME, logdir='log'):
    # make logdir if not existant
    try:
        os.makedirs(logdir)
    except OSError:
        pass
    #paths2process = deque()
    paths2process = OrderedDict()
    # watch only today's folder
    path_re = re.compile("(%s%s)/?$" % (topdir, check_ptrn))
    i = inotify.adapters.InotifyTree(topdir.encode())#, mask=MASK)
    for event in i.event_gen():
        if event is not None:
            (header, type_names, watch_path, filename) = event
            _LOGGER.info("WD=(%d) MASK=(%d) COOKIE=(%d) LEN=(%d) MASK->NAMES=%s"
                         " WATCH-PATH=[%s] FILENAME=[%s]",
                         header.wd, header.mask, header.cookie, header.len,
                         type_names, watch_path.decode('utf-8'),
                         filename.decode('utf-8'))
            if (header.mask == MASK_NEWDIR
                and path_re.match(watch_path.decode('utf-8'))):
                # we got our directory, now let's do something on it
                newpath2process = op.join(watch_path, filename).decode('utf-8')
                #paths2process.append(newpath2process)
                # update time
                paths2process[newpath2process] = time.time()
                print(newpath2process, time.time())
            # check if we need to update the time
            for path in paths2process.keys():
                if path in watch_path.decode('utf-8'):
                    paths2process[path] = time.time()
                    print("Updating {0}: {1}".format(path, paths2process[path]))

        # check if there's anything to process
        process(paths2process, db, wait=wait, logdir=logdir)


def parse_args():
    import argparse
    parser = argparse.ArgumentParser(
        prog='monitor.py',
        description=('Small monitoring script to detect new directories and '
                     'process them'),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('path',
        help='Which directory to monitor')
    parser.add_argument('--check_ptrn', '-p',
        help='regexp pattern for which subdirectories to check',
        default='/20../../..')
    parser.add_argument('--database', '-d',
        help='database location',
        default='database.json')
    parser.add_argument('--wait_time', '-w',
        help='After how long should we start processing datasets? (in seconds)',
        default=86400,
        type=float)
    parser.add_argument('--logdir', '-l',
        help='Where should we save the logs?',
        default='log')
    return parser.parse_args()


def main():
    parsed = parse_args()
    print('Got {0}'.format(parsed))
    # open database
    db = TinyDB(parsed.database, default_table='heudiconv')
    monitor(parsed.path, parsed.check_ptrn, db,
            wait=parsed.wait_time, logdir=parsed.logdir)


if __name__ == '__main__':
    main()



import logging
import os.path as op
import sys

from . import __version__, __packagename__
from .bids import populate_bids_templates, tuneup_bids_json_files
from .convert import prep_conversion
from .due import due, Doi
from .parser import get_study_sessions
from .queue import queue_conversion
from .utils import anonymize_sid, load_heuristic, treat_infofile, SeqInfo

lgr = logging.getLogger(__name__)


INIT_MSG = "Running {packname} version {version} latest {latest}".format


def is_interactive():
    """Return True if all in/outs are tty"""
    # TODO: check on windows if hasattr check would work correctly and add value:
    return sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty()


def setup_exceptionhook():
    """
    Overloads default sys.excepthook with our exceptionhook handler.

    If interactive, our exceptionhook handler will invoke pdb.post_mortem;
    if not interactive, then invokes default handler.
    """
    def _pdb_excepthook(type, value, tb):
        if is_interactive():
            import traceback
            import pdb
            traceback.print_exception(type, value, tb)
            # print()
            pdb.post_mortem(tb)
        else:
            lgr.warning(
              "We cannot setup exception hook since not in interactive mode")

    sys.excepthook = _pdb_excepthook


def process_extra_commands(outdir, command, files, dicom_dir_template,
                           heuristic, session, subjs, grouping):
    """
    Perform custom command instead of regular operations. Supported commands:
    ['treat-json', 'ls', 'populate-templates']

    Parameters
    ----------
    outdir : str
        Output directory
    command : {'treat-json', 'ls', 'populate-templates'}
        Heudiconv command to run
    files : list of str
        List of files
    dicom_dir_template : str
        Location of dicomdir that can be indexed with subject id
        {subject} and session {session}. Tarballs (can be compressed)
        are supported in addition to directory. All matching tarballs
        for a subject are extracted and their content processed in a
        single pass. If multiple tarballs are found, each is assumed to
        be a separate session and the 'session' argument is ignored.
    heuristic : str
        Path to heuristic file or name of builtin heuristic.
    session : str
        Session identifier
    subjs : list of str
        List of subject identifiers
    grouping : {'studyUID', 'accession_number', 'all', 'custom'}
        How to group dicoms.
    """
    if command == 'treat-jsons':
        for f in files:
            treat_infofile(f)
    elif command == 'ls':
        ensure_heuristic_arg(heuristic)
        heuristic = load_heuristic(heuristic)
        heuristic_ls = getattr(heuristic, 'ls', None)
        for f in files:
            study_sessions = get_study_sessions(
                dicom_dir_template, [f], heuristic, outdir,
                session, subjs, grouping=grouping)
            print(f)
            for study_session, sequences in study_sessions.items():
                suf = ''
                if heuristic_ls:
                    suf += heuristic_ls(study_session, sequences)
                print(
                    "\t%s %d sequences%s"
                    % (str(study_session), len(sequences), suf)
                )
    elif command == 'populate-templates':
        ensure_heuristic_arg(heuristic)
        heuristic = load_heuristic(heuristic)
        for f in files:
            populate_bids_templates(f, getattr(heuristic, 'DEFAULT_FIELDS', {}))
    elif command == 'sanitize-jsons':
        tuneup_bids_json_files(files)
    elif command == 'heuristics':
        from .utils import get_known_heuristics_with_descriptions
        for name_desc in get_known_heuristics_with_descriptions().items():
            print("- %s: %s" % name_desc)
    elif command == 'heuristic-info':
        ensure_heuristic_arg(heuristic)
        from .utils import get_heuristic_description
        print(get_heuristic_description(heuristic, full=True))
    else:
        raise ValueError("Unknown command %s" % command)
    return


def ensure_heuristic_arg(heuristic=None):
    """
    Check that the heuristic argument was provided.
    """
    from .utils import get_known_heuristic_names
    if not heuristic:
        raise ValueError("Specify heuristic using -f. Known are: %s"
                         % ', '.join(get_known_heuristic_names()))


@due.dcite(
    Doi('10.5281/zenodo.1012598'),
    path='heudiconv',
    description='Flexible DICOM converter for organizing brain imaging data',
    version=__version__,
    cite_module=True)
def workflow(*, dicom_dir_template=None, files=None, subjs=None,
             converter='dcm2niix', outdir='.', locator=None, conv_outdir=None,
             anon_cmd=None, heuristic=None, with_prov=False, session=None,
             bids_options=None, overwrite=False, datalad=False, debug=False,
             command=None, grouping='studyUID', minmeta=False,
             random_seed=None, dcmconfig=None, queue=None, queue_args=None):
    """Run the HeuDiConv conversion workflow.

    Parameters
    ----------
    dicom_dir_template : str or None, optional
        Location of dicomdir that can be indexed with subject id
        {subject} and session {session}. Tarballs (can be compressed)
        are supported in addition to directory. All matching tarballs
        for a subject are extracted and their content processed in a
        single pass. If multiple tarballs are found, each is assumed to
        be a separate session and the 'session' argument is ignored.
        Mutually exclusive with 'files'. Default is None.
    files : list or None, optional
        Files (tarballs, dicoms) or directories containing files to
        process. Mutually exclusive with 'dicom_dir_template'. Default is None.
    subjs : list or None, optional
        List of subjects - required for dicom template. If not
        provided, DICOMS would first be "sorted" and subject IDs
        deduced by the heuristic. Default is None.
    converter : {'dcm2niix', None}, optional
        Tool to use for DICOM conversion. Setting to None disables
        the actual conversion step -- useful for testing heuristics.
        Default is None.
    outdir : str, optional
        Output directory for conversion setup (for further
        customization and future reference. This directory will refer
        to non-anonymized subject IDs.
        Default is '.' (current working directory).
    locator : str or 'unknown' or None, optional
        Study path under outdir. If provided, it overloads the value
        provided by the heuristic. If 'datalad=True', every
        directory within locator becomes a super-dataset thus
        establishing a hierarchy. Setting to "unknown" will skip that
        dataset. Default is None.
    conv_outdir : str or None, optional
        Output directory for converted files. By default this is
        identical to --outdir. This option is most useful in
        combination with 'anon_cmd'. Default is None.
    anon_cmd : str or None, optional
        Command to run to convert subject IDs used for DICOMs to
        anonymized IDs. Such command must take a single argument and
        return a single anonymized ID. Also see 'conv_outdir'. Default is None.
    heuristic : str or None, optional
        Name of a known heuristic or path to the Python script containing
        heuristic. Default is None.
    with_prov : bool, optional
        Store additional provenance information. Requires python-rdflib.
        Default is False.
    session : str or None, optional
        Session for longitudinal study_sessions. Default is None.
    bids_options : str or None, optional
        Flag for output into BIDS structure. Can also take BIDS-
        specific options, e.g., --bids notop. The only currently
        supported options is "notop", which skips creation of
        top-level BIDS files. This is useful when running in batch
        mode to prevent possible race conditions. Default is None.
    overwrite : bool, optional
        Overwrite existing converted files. Default is False.
    datalad : bool, optional
        Store the entire collection as DataLad dataset(s). Small files
        will be committed directly to git, while large to annex. New
        version (6) of annex repositories will be used in a "thin"
        mode so it would look to mortals as just any other regular
        directory (i.e. no symlinks to under .git/annex). For now just
        for BIDS mode. Default is False.
    debug : bool, optional
        Do not catch exceptions and show exception traceback. Default is False.
    command : {'heuristics', 'heuristic-info', 'ls', 'populate-templates',
               'sanitize-jsons', 'treat-jsons', None}, optional
        Custom action to be performed on provided files instead of regular
        operation. Default is None.
    grouping : {'studyUID', 'accession_number', 'all', 'custom'}, optional
        How to group dicoms. Default is 'studyUID'.
    minmeta : bool, optional
        Exclude dcmstack meta information in sidecar jsons. Default is False.
    random_seed : int or None, optional
        Random seed to initialize RNG. Default is None.
    dcmconfig : str or None, optional
        JSON file for additional dcm2niix configuration. Default is None.
    queue : {'SLURM', None}, optional
        Batch system to submit jobs in parallel. Default is None.
        If set, will cause scheduling of conversion and return without performing
        any further action.
    queue_args : str or None, optional
        Additional queue arguments passed as single string of space-separated
        Argument=Value pairs. Default is None.

    Notes
    -----
    All parameters in this function must be called as keyword arguments.
    """

    # To be done asap so anything random is deterministic
    if random_seed is not None:
        import random
        random.seed(random_seed)
        import numpy
        numpy.random.seed(random_seed)
    # Ensure only supported bids options are passed
    if debug:
        lgr.setLevel(logging.DEBUG)
    # Should be possible but only with a single subject -- will be used to
    # override subject deduced from the DICOMs
    if files and subjs and len(subjs) > 1:
        raise ValueError(
            "Unable to processes multiple `--subjects` with files"
        )

    if debug:
        setup_exceptionhook()

    # Deal with provided files or templates
    # pre-process provided list of files and possibly sort into groups/sessions
    # Group files per each study/sid/session

    outdir = op.abspath(outdir)

    try:
        import etelemetry
        latest = etelemetry.get_project("nipy/heudiconv")
    except Exception as e:
        lgr.warning("Could not check for version updates: %s", str(e))
        latest = {"version": 'Unknown'}

    lgr.info(INIT_MSG(packname=__packagename__,
                      version=__version__,
                      latest=latest["version"]))

    if command:
        process_extra_commands(outdir, command, files, dicom_dir_template,
                               heuristic, session, subjs, grouping)
        return
    #
    # Load heuristic -- better do it asap to make sure it loads correctly
    #
    if not heuristic:
        raise RuntimeError("No heuristic specified - add to arguments and rerun")

    if queue:
        lgr.info("Queuing %s conversion", queue)
        iterarg, iterables = ("files", len(files)) if files else \
                             ("subjects", len(subjs))
        queue_conversion(queue, iterarg, iterables, queue_args)
        return

    heuristic = load_heuristic(heuristic)

    study_sessions = get_study_sessions(dicom_dir_template, files,
                                        heuristic, outdir, session,
                                        subjs, grouping=grouping)

    # extract tarballs, and replace their entries with expanded lists of files
    # TODO: we might need to sort so sessions are ordered???
    lgr.info("Need to process %d study sessions", len(study_sessions))

    # processed_studydirs = set()

    locator_manual, session_manual = locator, session
    for (locator, session, sid), files_or_seqinfo in study_sessions.items():

        # Allow for session to be overloaded from command line
        if session_manual is not None:
            session = session_manual
        if locator_manual is not None:
            locator = locator_manual
        if not len(files_or_seqinfo):
            raise ValueError("nothing to process?")
        # that is how life is ATM :-/ since we don't do sorting if subj
        # template is provided
        if isinstance(files_or_seqinfo, dict):
            assert(isinstance(list(files_or_seqinfo.keys())[0], SeqInfo))
            dicoms = None
            seqinfo = files_or_seqinfo
        else:
            dicoms = files_or_seqinfo
            seqinfo = None

        if locator == 'unknown':
            lgr.warning("Skipping unknown locator dataset")
            continue

        anon_sid = anonymize_sid(sid, anon_cmd) if anon_cmd else None
        if anon_cmd:
            lgr.info('Anonymized {} to {}'.format(sid, anon_sid))

        study_outdir = op.join(outdir, locator or '')
        anon_outdir = conv_outdir or outdir
        anon_study_outdir = op.join(anon_outdir, locator or '')

        if datalad:
            from .external.dlad import prepare_datalad
            dlad_sid = sid if not anon_sid else anon_sid
            dl_msg = prepare_datalad(anon_study_outdir, anon_outdir, dlad_sid,
                                     session, seqinfo, dicoms,
                                     bids_options)

        lgr.info("PROCESSING STARTS: {0}".format(
            str(dict(subject=sid, outdir=study_outdir, session=session))))

        prep_conversion(sid,
                        dicoms,
                        study_outdir,
                        heuristic,
                        converter=converter,
                        anon_sid=anon_sid,
                        anon_outdir=anon_study_outdir,
                        with_prov=with_prov,
                        ses=session,
                        bids_options=bids_options,
                        seqinfo=seqinfo,
                        min_meta=minmeta,
                        overwrite=overwrite,
                        dcmconfig=dcmconfig,
                        grouping=grouping,)

        lgr.info("PROCESSING DONE: {0}".format(
            str(dict(subject=sid, outdir=study_outdir, session=session))))

        if datalad:
            from .external.dlad import add_to_datalad
            msg = "Converted subject %s" % dl_msg
            # TODO:  whenever propagate to supers work -- do just
            # ds.save(msg=msg)
            #  also in batch mode might fail since we have no locking ATM
            #  and theoretically no need actually to save entire study
            #  we just need that
            add_to_datalad(outdir, study_outdir, msg, bids_options)

    # if bids:
    #     # Let's populate BIDS templates for folks to take care about
    #     for study_outdir in processed_studydirs:
    #         populate_bids_templates(study_outdir)
    #
    # TODO: record_collection of the sid/session although that information
    # is pretty much present in .heudiconv/SUBJECT/info so we could just poke there



import os.path as op
import json

import pytest

from heudiconv.external.pydicom import dcm
from heudiconv.cli.run import main as runner
from heudiconv.convert import nipype_convert
from heudiconv.dicoms import parse_private_csa_header, embed_dicom_and_nifti_metadata
from .utils import (
    assert_cwd_unchanged,
    TESTS_DATA_PATH,
)

# Public: Private DICOM tags
DICOM_FIELDS_TO_TEST = {
    'ProtocolName': 'tProtocolName'
}

def test_private_csa_header(tmpdir):
    dcm_file = op.join(TESTS_DATA_PATH, 'axasc35.dcm')
    dcm_data = dcm.read_file(dcm_file, stop_before_pixels=True)
    for pub, priv in DICOM_FIELDS_TO_TEST.items():
        # ensure missing public tag
        with pytest.raises(AttributeError):
            dcm.pub
        # ensure private tag is found
        assert parse_private_csa_header(dcm_data, pub, priv) != ''
        # and quickly run heudiconv with no conversion
        runner(['--files', dcm_file, '-c' 'none', '-f', 'reproin'])


@assert_cwd_unchanged(ok_to_chdir=True)  # so we cd back after tmpdir.chdir
def test_embed_dicom_and_nifti_metadata(tmpdir):
    """Test dcmstack's additional fields"""
    tmpdir.chdir()
    # set up testing files
    dcmfiles = [op.join(TESTS_DATA_PATH, 'axasc35.dcm')]
    infofile = 'infofile.json'

    out_prefix = str(tmpdir / "nifti")
    # 1) nifti does not exist -- no longer supported
    with pytest.raises(NotImplementedError):
        embed_dicom_and_nifti_metadata(dcmfiles, out_prefix + '.nii.gz', infofile, None)

    # we should produce nifti using our "standard" ways
    nipype_out, prov_file = nipype_convert(
        dcmfiles, prefix=out_prefix, with_prov=False,
        bids_options=None, tmpdir=str(tmpdir))
    niftifile = nipype_out.outputs.converted_files

    assert op.exists(niftifile)
    # 2) nifti exists
    embed_dicom_and_nifti_metadata(dcmfiles, niftifile, infofile, None)
    assert op.exists(infofile)
    with open(infofile) as fp:
        out2 = json.load(fp)

    # 3) with existing metadata
    bids = {"existing": "data"}
    embed_dicom_and_nifti_metadata(dcmfiles, niftifile, infofile, bids)
    with open(infofile) as fp:
        out3 = json.load(fp)

    assert out3.pop("existing") == "data"
    assert out3 == out2



from functools import wraps
import os
import os.path as op
import sys

import heudiconv.heuristics


HEURISTICS_PATH = op.join(heudiconv.heuristics.__path__[0])
TESTS_DATA_PATH = op.join(op.dirname(__file__), 'data')

import logging
lgr = logging.getLogger(__name__)


def gen_heudiconv_args(datadir, outdir, subject, heuristic_file,
                       anon_cmd=None, template=None, xargs=None):
    heuristic = op.realpath(op.join(HEURISTICS_PATH, heuristic_file))

    if template:
        # use --dicom_dir_template
        args = ["-d", op.join(datadir, template)]
    else:
        args = ["--files", datadir]

    args.extend([
            "-c", "dcm2niix",
            "-o", outdir,
            "-s", subject,
            "-f", heuristic,
            "--bids",
            "--minmeta",]
            )
    if anon_cmd:
        args += ["--anon-cmd", op.join(op.dirname(__file__), anon_cmd), "-a", outdir]
    if xargs:
        args += xargs

    return args


def fetch_data(tmpdir, dataset, getpath=None):
    """
    Utility function to interface with datalad database.
    Performs datalad `install` and datalad `get` operations.

    Parameters
    ----------
    tmpdir : str
        directory to temporarily store data
    dataset : str
        dataset path from `http://datasets-tests.datalad.org`
    getpath : str [optional]
        exclusive path to get

    Returns
    -------
    targetdir : str
        directory with installed dataset
    """
    from datalad import api
    targetdir = op.join(tmpdir, op.basename(dataset))
    ds = api.install(path=targetdir,
                source='http://datasets-tests.datalad.org/{}'.format(dataset))

    getdir = targetdir + (op.sep + getpath if getpath is not None else '')
    ds.get(getdir)
    return targetdir


def assert_cwd_unchanged(ok_to_chdir=False):
    """Decorator to test whether the current working directory remains unchanged

    Provenance: based on the one in datalad, but simplified.

    Parameters
    ----------
    ok_to_chdir: bool, optional
      If True, allow to chdir, so this decorator would not then raise exception
      if chdir'ed but only return to original directory
    """

    def decorator(func=None):  # =None to avoid pytest treating it as a fixture
        @wraps(func)
        def newfunc(*args, **kwargs):
            cwd_before = os.getcwd()
            exc = None
            try:
                return func(*args, **kwargs)
            except Exception as exc_:
                exc = exc_
            finally:
                try:
                    cwd_after = os.getcwd()
                except OSError as e:
                    lgr.warning("Failed to getcwd: %s" % e)
                    cwd_after = None

                if cwd_after != cwd_before:
                    os.chdir(cwd_before)
                    if not ok_to_chdir:
                        lgr.warning(
                            "%s changed cwd to %s. Mitigating and changing back to %s"
                            % (func, cwd_after, cwd_before))
                        # If there was already exception raised, we better reraise
                        # that one since it must be more important, so not masking it
                        # here with our assertion
                        if exc is None:
                            assert cwd_before == cwd_after, \
                                     "CWD changed from %s to %s" % (cwd_before, cwd_after)

                if exc is not None:
                    raise exc
        return newfunc

    return decorator



import inspect
import os

import os.path as op
import logging
from glob import glob
from os import path as op

from ..utils import create_file_if_missing

lgr = logging.getLogger(__name__)

from ..info import MIN_DATALAD_VERSION as MIN_VERSION


def prepare_datalad(studydir, outdir, sid, session, seqinfo, dicoms, bids):
    """ Prepare data for datalad """
    from datalad.api import Dataset
    datalad_msg_suf = ' %s' % sid
    if session:
        datalad_msg_suf += ", session %s" % session
    if seqinfo:
        datalad_msg_suf += ", %d sequences" % len(seqinfo)
    datalad_msg_suf += ", %d dicoms" % (len(sum(seqinfo.values(), []))
                                        if seqinfo else len(dicoms))
    ds = Dataset(studydir)
    if not op.exists(outdir) or not ds.is_installed():
        add_to_datalad(outdir, studydir,
                       msg="Preparing for %s" % datalad_msg_suf,
                       bids=bids)
    return datalad_msg_suf


def add_to_datalad(topdir, studydir, msg, bids):
    """Do all necessary preparations (if were not done before) and save
    """
    import datalad.api as dl
    from datalad.api import Dataset
    from datalad.support.annexrepo import AnnexRepo
    from datalad.support.external_versions import external_versions
    assert external_versions['datalad'] >= MIN_VERSION, (
        "Need datalad >= {}".format(MIN_VERSION))  # add to reqs


    studyrelpath = op.relpath(studydir, topdir)
    assert not studyrelpath.startswith(op.pardir)  # so we are under
    # now we need to test and initiate a DataLad dataset all along the path
    curdir_ = topdir
    superds = None
    subdirs = [''] + [d for d in studyrelpath.split(op.sep) if d != os.curdir]
    for isubdir, subdir in enumerate(subdirs):
        curdir_ = op.join(curdir_, subdir)
        ds = Dataset(curdir_)
        if not ds.is_installed():
            lgr.info("Initiating %s", ds)
            # would require annex > 20161018 for correct operation on annex v6
            # need to add .gitattributes first anyways
            ds_ = dl.create(curdir_, dataset=superds,
                         force=True,
                         # initiate annex only at the bottom repository
                         no_annex=isubdir<(len(subdirs)-1),
                         fake_dates=True,
                         # shared_access='all',
                         )
            assert ds == ds_
        assert ds.is_installed()
        superds = ds

    # TODO: we need a helper (in DataLad ideally) to ease adding such
    # specifications
    gitattributes_path = op.join(studydir, '.gitattributes')
    # We will just make sure that all our desired rules are present in it
    desired_attrs = """\
* annex.largefiles=(largerthan=100kb)
*.json annex.largefiles=nothing
*.txt annex.largefiles=nothing
*.tsv annex.largefiles=nothing
*.nii.gz annex.largefiles=anything
*.tgz annex.largefiles=anything
*_scans.tsv annex.largefiles=anything
"""
    if op.exists(gitattributes_path):
        with open(gitattributes_path, 'rb') as f:
            known_attrs = [line.decode('utf-8').rstrip() for line in f.readlines()]
    else:
        known_attrs = []
    for attr in desired_attrs.split('\n'):
        if attr not in known_attrs:
            known_attrs.append(attr)
    with open(gitattributes_path, 'wb') as f:
        f.write('\n'.join(known_attrs).encode('utf-8'))


    # ds might have memories of having ds.repo GitRepo
    superds = Dataset(topdir)
    assert op.realpath(ds.path) == op.realpath(studydir)
    assert isinstance(ds.repo, AnnexRepo)
    # Add doesn't have all the options of save such as msg and supers
    ds.save(path=['.gitattributes'], message="Custom .gitattributes", to_git=True)
    dsh = dsh_path = None
    if op.lexists(op.join(ds.path, '.heudiconv')):
        dsh_path = op.join(ds.path, '.heudiconv')
        dsh = Dataset(dsh_path)
        if not dsh.is_installed():
            # Previously we did not have it as a submodule, and since no
            # automagic migration is implemented, we just need to check first
            # if any path under .heudiconv is already under git control
            if any(x.startswith('.heudiconv/') for x in ds.repo.get_files()):
                lgr.warning("%s has .heudiconv not as a submodule from previous"
                            " versions of heudiconv. No automagic migration is "
                            "yet provided", ds)
            else:
                dsh = ds.create(path='.heudiconv',
                                force=True,
                                # shared_access='all'
                                )
        # Since .heudiconv could contain sensitive information
        # we place all files under annex and then add
        if create_file_if_missing(op.join(dsh_path, '.gitattributes'),
                                  """* annex.largefiles=anything"""):
            ds.save('.heudiconv/.gitattributes',
                   to_git=True,
                   message="Added gitattributes to place all .heudiconv content"
                           " under annex")
    ds.save('.', recursive=True
           # not in effect! ?
           #annex_add_opts=['--include-dotfiles']
           )

    # TODO: filter for only changed files?
    # Provide metadata for sensitive information
    mark_sensitive(ds, 'sourcedata')
    mark_sensitive(ds, '*_scans.tsv')  # top level
    mark_sensitive(ds, '*/*_scans.tsv')  # within subj
    mark_sensitive(ds, '*/*/*_scans.tsv')  # within sess/subj
    mark_sensitive(ds, '*/anat')  # within subj
    mark_sensitive(ds, '*/*/anat')  # within ses/subj
    if dsh_path:
        mark_sensitive(ds, '.heudiconv')  # entire .heudiconv!
    superds.save(path=ds.path, message=msg, recursive=True)

    assert not ds.repo.dirty
    # TODO:  they are still appearing as native annex symlinked beasts
    """
    TODOs:
    it needs
    - unlock  (thin will be in effect)
    - save/commit (does modechange 120000 => 100644
    - could potentially somehow automate that all:
      http://git-annex.branchable.com/tips/automatically_adding_metadata/
    - possibly even make separate sub-datasets for originaldata, derivatives ?
    """


def mark_sensitive(ds, path_glob):
    """

    Parameters
    ----------
    ds : Dataset to operate on
    path_glob : str
      glob of the paths within dataset to work on

    Returns
    -------
    None
    """
    paths = glob(op.join(ds.path, path_glob))
    if not paths:
        return
    lgr.debug("Marking %d files with distribution-restrictions field",
              len(paths))
    # set_metadata can be a bloody generator
    res = ds.repo.set_metadata(
        paths,
        init=dict([('distribution-restrictions', 'sensitive')]),
        recursive=True)
    if inspect.isgenerator(res):
        res = list(res)



# set logger handler
import logging
import os
from .info import (__version__, __packagename__)

# Rudimentary logging support.
lgr = logging.getLogger(__name__)
logging.basicConfig(
    format='%(levelname)s: %(message)s',
    level=getattr(logging, os.environ.get('HEUDICONV_LOG_LEVEL', 'INFO'))
)
lgr.debug("Starting the abomination")  # just to "run-test" logging



# dicom operations
import os
import os.path as op
import logging
from collections import OrderedDict
import tarfile

from .external.pydicom import dcm
from .utils import (
    get_typed_attr,
    load_json,
    save_json,
    SeqInfo,
    set_readonly,
)

import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # suppress warning
    import nibabel.nicom.dicomwrappers as dw

lgr = logging.getLogger(__name__)
total_files = 0


def create_seqinfo(mw, series_files, series_id):
    """Generate sequence info

    Parameters
    ----------
    mw: MosaicWrapper
    series_files: list
    series_id: str
    """
    dcminfo = mw.dcm_data
    accession_number = dcminfo.get('AccessionNumber')

    # TODO: do not group echoes by default
    size = list(mw.image_shape) + [len(series_files)]
    if len(size) < 4:
        size.append(1)

    # parse DICOM for seqinfo fields
    TR = get_typed_attr(dcminfo, "RepetitionTime", float, -1000) / 1000
    TE = get_typed_attr(dcminfo, "EchoTime", float, -1)
    refphys = get_typed_attr(dcminfo, "ReferringPhysicianName", str, "")
    image_type = get_typed_attr(dcminfo, "ImageType", tuple, ())
    is_moco = 'MOCO' in image_type
    series_desc = get_typed_attr(dcminfo, "SeriesDescription", str, "")

    if dcminfo.get([0x18, 0x24]):
        # GE and Philips
        sequence_name = dcminfo[0x18, 0x24].value
    elif dcminfo.get([0x19, 0x109c]):
        # Siemens
        sequence_name = dcminfo[0x19, 0x109c].value
    else:
        sequence_name = ""

    # initialized in `group_dicoms_to_seqinfos`
    global total_files
    total_files += len(series_files)

    seqinfo = SeqInfo(
        total_files_till_now=total_files,
        example_dcm_file=op.basename(series_files[0]),
        series_id=series_id,
        dcm_dir_name=op.basename(op.dirname(series_files[0])),
        series_files=len(series_files),
        unspecified="",
        dim1=size[0],
        dim2=size[1],
        dim3=size[2],
        dim4=size[3],
        TR=TR,
        TE=TE,
        protocol_name=dcminfo.ProtocolName,
        is_motion_corrected=is_moco,
        is_derived='derived' in [x.lower() for x in image_type],
        patient_id=dcminfo.get('PatientID'),
        study_description=dcminfo.get('StudyDescription'),
        referring_physician_name=refphys,
        series_description=series_desc,
        sequence_name=sequence_name,
        image_type=image_type,
        accession_number=accession_number,
        # For demographics to populate BIDS participants.tsv
        patient_age=dcminfo.get('PatientAge'),
        patient_sex=dcminfo.get('PatientSex'),
        date=dcminfo.get('AcquisitionDate'),
        series_uid=dcminfo.get('SeriesInstanceUID')
    )
    return seqinfo


def validate_dicom(fl, dcmfilter):
    """
    Parse DICOM attributes. Returns None if not valid.
    """
    mw = dw.wrapper_from_file(fl, force=True, stop_before_pixels=True)
    # clean series signature
    for sig in ('iop', 'ICE_Dims', 'SequenceName'):
        try:
            del mw.series_signature[sig]
        except KeyError:
            pass
    # Workaround for protocol name in private siemens csa header
    if not getattr(mw.dcm_data, 'ProtocolName', '').strip():
        mw.dcm_data.ProtocolName = parse_private_csa_header(
            mw.dcm_data, 'ProtocolName', 'tProtocolName'
        ) if mw.is_csa else ''
    try:
        series_id = (
            int(mw.dcm_data.SeriesNumber), mw.dcm_data.ProtocolName
        )
    except AttributeError as e:
        lgr.warning(
            'Ignoring %s since not quite a "normal" DICOM: %s', fl, e
        )
        return
    if dcmfilter is not None and dcmfilter(mw.dcm_data):
        lgr.warning("Ignoring %s because of DICOM filter", fl)
        return
    if mw.dcm_data[0x0008, 0x0016].repval in (
        'Raw Data Storage',
        'GrayscaleSoftcopyPresentationStateStorage'
    ):
        return
    try:
        file_studyUID = mw.dcm_data.StudyInstanceUID
    except AttributeError:
        lgr.info("File {} is missing any StudyInstanceUID".format(fl))
        file_studyUID = None
    return mw, series_id, file_studyUID


def group_dicoms_into_seqinfos(files, grouping, file_filter=None,
                               dcmfilter=None, flatten=False,
                               custom_grouping=None):
    """Process list of dicoms and return seqinfo and file group
    `seqinfo` contains per-sequence extract of fields from DICOMs which
    will be later provided into heuristics to decide on filenames

    Parameters
    ----------
    files : list of str
      List of files to consider
    grouping : {'studyUID', 'accession_number', 'all', 'custom'}
      How to group DICOMs for conversion. If 'custom', see `custom_grouping`
      parameter.
    file_filter : callable, optional
      Applied to each item of filenames. Should return True if file needs to be
      kept, False otherwise.
    dcmfilter : callable, optional
      If called on dcm_data and returns True, it is used to set series_id
    flatten : bool, optional
      Creates a flattened `seqinfo` with corresponding DICOM files. True when
      invoked with `dicom_dir_template`.
    custom_grouping: str or callable, optional
     grouping key defined within heuristic. Can be a string of a
     DICOM attribute, or a method that handles more complex groupings.


    Returns
    -------
    seqinfo : list of list
      `seqinfo` is a list of info entries per each sequence (some entry
      there defines a key for `filegrp`)
    filegrp : dict
      `filegrp` is a dictionary with files groupped per each sequence
    """
    allowed_groupings = ['studyUID', 'accession_number', 'all', 'custom']
    if grouping not in allowed_groupings:
        raise ValueError('I do not know how to group by {0}'.format(grouping))
    per_studyUID = grouping == 'studyUID'
    # per_accession_number = grouping == 'accession_number'
    lgr.info("Analyzing %d dicoms", len(files))

    groups = [[], []]
    mwgroup = []
    studyUID = None

    if file_filter:
        nfl_before = len(files)
        files = list(filter(file_filter, files))
        nfl_after = len(files)
        lgr.info('Filtering out {0} dicoms based on their filename'.format(
            nfl_before-nfl_after))

    if grouping == 'custom':
        if custom_grouping is None:
            raise RuntimeError("Custom grouping is not defined in heuristic")
        if callable(custom_grouping):
            return custom_grouping(files, dcmfilter, SeqInfo)
        grouping = custom_grouping
        study_customgroup = None

    removeidx = []
    for idx, filename in enumerate(files):
        mwinfo = validate_dicom(filename, dcmfilter)
        if mwinfo is None:
            removeidx.append(idx)
            continue
        mw, series_id, file_studyUID = mwinfo
        if per_studyUID:
            series_id = series_id + (file_studyUID,)

        if flatten:
            if per_studyUID:
                if studyUID is None:
                    studyUID = file_studyUID
                assert studyUID == file_studyUID, (
                    "Conflicting study identifiers found [{}, {}]."
                    .format(studyUID, file_studyUID)
                )
            elif custom_grouping:
                file_customgroup = mw.dcm_data.get(grouping)
                if study_customgroup is None:
                    study_customgroup = file_customgroup
                assert study_customgroup == file_customgroup, (
                    "Conflicting {0} found: [{1}, {2}]"
                    .format(grouping, study_customgroup, file_customgroup)
                )

        ingrp = False
        # check if same series was already converted
        for idx in range(len(mwgroup)):
            if mw.is_same_series(mwgroup[idx]):
                if grouping != 'all':
                    assert (
                        mwgroup[idx].dcm_data.get('StudyInstanceUID') == file_studyUID
                    ), "Same series found for multiple different studies"
                ingrp = True
                series_id = (
                    mwgroup[idx].dcm_data.SeriesNumber,
                    mwgroup[idx].dcm_data.ProtocolName
                )
                if per_studyUID:
                    series_id = series_id + (file_studyUID,)
                groups[0].append(series_id)
                groups[1].append(idx)

        if not ingrp:
            mwgroup.append(mw)
            groups[0].append(series_id)
            groups[1].append(len(mwgroup) - 1)

    group_map = dict(zip(groups[0], groups[1]))

    if removeidx:
        # remove non DICOMS from files
        for idx in sorted(removeidx, reverse=True):
            del files[idx]

    seqinfos = OrderedDict()
    # for the next line to make any sense the series_id needs to
    # be sortable in a way that preserves the series order
    for series_id, mwidx in sorted(group_map.items()):
        mw = mwgroup[mwidx]
        series_files = [files[i] for i, s in enumerate(groups[0]) if s == series_id]
        if per_studyUID:
            studyUID = series_id[2]
            series_id = series_id[:2]
        series_id = '-'.join(map(str, series_id))
        if mw.image_shape is None:
            # this whole thing has no image data (maybe just PSg DICOMs)
            # nothing to see here, just move on
            continue
        seqinfo = create_seqinfo(mw, series_files, series_id)

        if per_studyUID:
            key = studyUID
        elif grouping == 'accession_number':
            key = mw.dcm_data.get("AccessionNumber")
        elif grouping == 'all':
            key = 'all'
        elif custom_grouping:
            key = mw.dcm_data.get(custom_grouping)
        else:
            key = ''
        lgr.debug("%30s %30s %27s %27s %5s nref=%-2d nsrc=%-2d %s" % (
            key,
            seqinfo.series_id,
            seqinfo.series_description,
            mw.dcm_data.ProtocolName,
            seqinfo.is_derived,
            len(mw.dcm_data.get('ReferencedImageSequence', '')),
            len(mw.dcm_data.get('SourceImageSequence', '')),
            seqinfo.image_type
        ))

        if not flatten:
            if key not in seqinfos:
                seqinfos[key] = OrderedDict()
            seqinfos[key][seqinfo] = series_files
        else:
            seqinfos[seqinfo] = series_files

    if per_studyUID:
        lgr.info("Generated sequence info for %d studies with %d entries total",
                 len(seqinfos), sum(map(len, seqinfos.values())))
    elif grouping == 'accession_number':
        lgr.info("Generated sequence info for %d accession numbers with %d "
                 "entries total", len(seqinfos), sum(map(len, seqinfos.values())))
    else:
        lgr.info("Generated sequence info with %d entries", len(seqinfos))
    return seqinfos


def get_dicom_series_time(dicom_list):
    """Get time in seconds since epoch from dicom series date and time
    Primarily to be used for reproducible time stamping
    """
    import time
    import calendar

    dicom = dcm.read_file(dicom_list[0], stop_before_pixels=True, force=True)
    dcm_date = dicom.SeriesDate  # YYYYMMDD
    dcm_time = dicom.SeriesTime  # HHMMSS.MICROSEC
    dicom_time_str = dcm_date + dcm_time.split('.', 1)[0]  # YYYYMMDDHHMMSS
    # convert to epoch
    return calendar.timegm(time.strptime(dicom_time_str, '%Y%m%d%H%M%S'))


def compress_dicoms(dicom_list, out_prefix, tempdirs, overwrite):
    """Archives DICOMs into a tarball

    Also tries to do it reproducibly, so takes the date for files
    and target tarball based on the series time (within the first file)

    Parameters
    ----------
    dicom_list : list of str
      list of dicom files
    out_prefix : str
      output path prefix, including the portion of the output file name
      before .dicom.tgz suffix
    tempdirs : object
      TempDirs object to handle multiple tmpdirs
    overwrite : bool
      Overwrite existing tarfiles

    Returns
    -------
    filename : str
      Result tarball
    """

    tmpdir = tempdirs(prefix='dicomtar')
    outtar = out_prefix + '.dicom.tgz'

    if op.exists(outtar) and not overwrite:
        lgr.info("File {} already exists, will not overwrite".format(outtar))
        return
    # tarfile encodes current time.time inside making those non-reproducible
    # so we should choose which date to use.
    # Solution from DataLad although ugly enough:

    dicom_list = sorted(dicom_list)
    dcm_time = get_dicom_series_time(dicom_list)

    def _assign_dicom_time(ti):
        # Reset the date to match the one of the last commit, not from the
        # filesystem since git doesn't track those at all
        ti.mtime = dcm_time
        return ti

    # poor man mocking since can't rely on having mock
    try:
        import time
        _old_time = time.time
        time.time = lambda: dcm_time
        if op.lexists(outtar):
            os.unlink(outtar)
        with tarfile.open(outtar, 'w:gz', dereference=True) as tar:
            for filename in dicom_list:
                outfile = op.join(tmpdir, op.basename(filename))
                if not op.islink(outfile):
                    os.symlink(op.realpath(filename), outfile)
                # place into archive stripping any lead directories and
                # adding the one corresponding to prefix
                tar.add(outfile,
                        arcname=op.join(op.basename(out_prefix),
                                        op.basename(outfile)),
                        recursive=False,
                        filter=_assign_dicom_time)
    finally:
        time.time = _old_time
        tempdirs.rmtree(tmpdir)

    return outtar


def embed_dicom_and_nifti_metadata(dcmfiles, niftifile, infofile, bids_info):
    """Embed metadata from nifti (affine etc) and dicoms into infofile (json)

    `niftifile` should exist. Its affine's orientation information is used while
    establishing new `NiftiImage` out of dicom stack and together with `bids_info`
    (if provided) is dumped into json `infofile`

    Parameters
    ----------
    dcmfiles
    niftifile
    infofile
    bids_info: dict
      Additional metadata to be embedded. `infofile` is overwritten if exists,
      so here you could pass some metadata which would overload (at the first
      level of the dict structure, no recursive fancy updates) what is obtained
      from nifti and dicoms

    """
    # imports for nipype
    import nibabel as nb
    import os.path as op
    import json
    import re
    from heudiconv.utils import save_json

    from heudiconv.external.dcmstack import ds
    stack = ds.parse_and_stack(dcmfiles, force=True).values()
    if len(stack) > 1:
        raise ValueError('Found multiple series')
    # may be odict now - iter to be safe
    stack = next(iter(stack))

    if not op.exists(niftifile):
        raise NotImplementedError(
            "%s does not exist. "
            "We are not producing new nifti files here any longer. "
            "Use dcm2niix directly or .convert.nipype_convert helper ."
            % niftifile
        )

    orig_nii = nb.load(niftifile)
    aff = orig_nii.affine
    ornt = nb.orientations.io_orientation(aff)
    axcodes = nb.orientations.ornt2axcodes(ornt)
    new_nii = stack.to_nifti(voxel_order=''.join(axcodes), embed_meta=True)
    meta_info = ds.NiftiWrapper(new_nii).meta_ext.to_json()
    meta_info = json.loads(meta_info)

    if bids_info:
        meta_info.update(bids_info)

    # write to outfile
    save_json(infofile, meta_info)


def embed_metadata_from_dicoms(bids_options, item_dicoms, outname, outname_bids,
                               prov_file, scaninfo, tempdirs, with_prov):
    """
    Enhance sidecar information file with more information from DICOMs

    Parameters
    ----------
    bids_options
    item_dicoms
    outname
    outname_bids
    prov_file
    scaninfo
    tempdirs
    with_prov

    Returns
    -------

    """
    from nipype import Node, Function
    tmpdir = tempdirs(prefix='embedmeta')

    # We need to assure that paths are absolute if they are relative
    item_dicoms = list(map(op.abspath, item_dicoms))

    embedfunc = Node(Function(input_names=['dcmfiles', 'niftifile', 'infofile',
                                           'bids_info',],
                              function=embed_dicom_and_nifti_metadata),
                     name='embedder')
    embedfunc.inputs.dcmfiles = item_dicoms
    embedfunc.inputs.niftifile = op.abspath(outname)
    embedfunc.inputs.infofile = op.abspath(scaninfo)
    embedfunc.inputs.bids_info = load_json(op.abspath(outname_bids)) if (bids_options is not None) else None
    embedfunc.base_dir = tmpdir
    cwd = os.getcwd()

    lgr.debug("Embedding into %s based on dicoms[0]=%s for nifti %s",
              scaninfo, item_dicoms[0], outname)
    try:
        if op.lexists(scaninfo):
            # TODO: handle annexed file case
            if not op.islink(scaninfo):
                set_readonly(scaninfo, False)
        res = embedfunc.run()
        set_readonly(scaninfo)
        if with_prov:
            g = res.provenance.rdf()
            g.parse(prov_file,
                    format='turtle')
            g.serialize(prov_file, format='turtle')
            set_readonly(prov_file)
    except Exception as exc:
        lgr.error("Embedding failed: %s", str(exc))
        os.chdir(cwd)

def parse_private_csa_header(dcm_data, public_attr, private_attr, default=None):
    """
    Parses CSA header in cases where value is not defined publicly

    Parameters
    ----------
    dcm_data : pydicom Dataset object
        DICOM metadata
    public_attr : string
        non-private DICOM attribute
    private_attr : string
        private DICOM attribute
    default (optional)
        default value if private_attr not found

    Returns
    -------
    val (default: empty string)
        private attribute value or default
    """
    # TODO: provide mapping to private_attr from public_attr
    from nibabel.nicom import csareader
    import dcmstack.extract as dsextract
    try:
        # TODO: test with attr besides ProtocolName
        csastr = csareader.get_csa_header(dcm_data, 'series')['tags']['MrPhoenixProtocol']['items'][0]
        csastr = csastr.replace("### ASCCONV BEGIN", "### ASCCONV BEGIN ### ")
        parsedhdr = dsextract.parse_phoenix_prot('MrPhoenixProtocol', csastr)
        val = parsedhdr[private_attr].replace(' ', '')
    except Exception as e:
        lgr.debug("Failed to parse CSA header: %s", str(e))
        val = default or ""
    return val



"""Test functions in heudiconv.convert module.
"""
import pytest

from heudiconv.convert import (update_complex_name,
                               update_multiecho_name,
                               update_uncombined_name)
from heudiconv.bids import BIDSError


def test_update_complex_name():
    """Unit testing for heudiconv.convert.update_complex_name(), which updates
    filenames with the part field if appropriate.
    """
    # Standard name update
    fn = 'sub-X_ses-Y_task-Z_run-01_sbref'
    metadata = {'ImageType': ['ORIGINAL', 'PRIMARY', 'P', 'MB', 'TE3', 'ND', 'MOSAIC']}
    suffix = 3
    out_fn_true = 'sub-X_ses-Y_task-Z_run-01_part-phase_sbref'
    out_fn_test = update_complex_name(metadata, fn, suffix)
    assert out_fn_test == out_fn_true
    # Catch an unsupported type and *do not* update
    fn = 'sub-X_ses-Y_task-Z_run-01_phase'
    out_fn_test = update_complex_name(metadata, fn, suffix)
    assert out_fn_test == fn
    # Data type is missing from metadata so use suffix
    fn = 'sub-X_ses-Y_task-Z_run-01_sbref'
    metadata = {'ImageType': ['ORIGINAL', 'PRIMARY', 'MB', 'TE3', 'ND', 'MOSAIC']}
    out_fn_true = 'sub-X_ses-Y_task-Z_run-01_part-3_sbref'
    out_fn_test = update_complex_name(metadata, fn, suffix)
    assert out_fn_test == out_fn_true
    # Catch existing field with value that *does not match* metadata
    # and raise Exception
    fn = 'sub-X_ses-Y_task-Z_run-01_part-mag_sbref'
    metadata = {'ImageType': ['ORIGINAL', 'PRIMARY', 'P', 'MB', 'TE3', 'ND', 'MOSAIC']}
    suffix = 3
    with pytest.raises(BIDSError):
        assert update_complex_name(metadata, fn, suffix)


def test_update_multiecho_name():
    """Unit testing for heudiconv.convert.update_multiecho_name(), which updates
    filenames with the echo field if appropriate.
    """
    # Standard name update
    fn = 'sub-X_ses-Y_task-Z_run-01_bold'
    metadata = {'EchoTime': 0.01,
                'EchoNumber': 1}
    echo_times = [0.01, 0.02, 0.03]
    out_fn_true = 'sub-X_ses-Y_task-Z_run-01_echo-1_bold'
    out_fn_test = update_multiecho_name(metadata, fn, echo_times)
    assert out_fn_test == out_fn_true
    # EchoNumber field is missing from metadata, so use echo_times
    metadata = {'EchoTime': 0.01}
    out_fn_test = update_multiecho_name(metadata, fn, echo_times)
    assert out_fn_test == out_fn_true
    # Catch an unsupported type and *do not* update
    fn = 'sub-X_ses-Y_task-Z_run-01_phasediff'
    out_fn_test = update_multiecho_name(metadata, fn, echo_times)
    assert out_fn_test == fn


def test_update_uncombined_name():
    """Unit testing for heudiconv.convert.update_uncombined_name(), which updates
    filenames with the ch field if appropriate.
    """
    # Standard name update
    fn = 'sub-X_ses-Y_task-Z_run-01_bold'
    metadata = {'CoilString': 'H1'}
    channel_names = ['H1', 'H2', 'H3', 'HEA;HEP']
    out_fn_true = 'sub-X_ses-Y_task-Z_run-01_ch-01_bold'
    out_fn_test = update_uncombined_name(metadata, fn, channel_names)
    assert out_fn_test == out_fn_true
    # CoilString field has no number in it
    metadata = {'CoilString': 'HEA;HEP'}
    out_fn_true = 'sub-X_ses-Y_task-Z_run-01_ch-04_bold'
    out_fn_test = update_uncombined_name(metadata, fn, channel_names)
    assert out_fn_test == out_fn_true



from ..dlad import mark_sensitive
from ...utils import create_tree

import pytest

dl = pytest.importorskip('datalad.api')


def test_mark_sensitive(tmpdir):
    ds = dl.Dataset(str(tmpdir)).create(force=True)
    create_tree(
        str(tmpdir),
        {
            'f1': 'd1',
            'f2': 'd2',
            'g1': 'd3',
            'g2': 'd1',
         }
    )
    ds.save('.')
    mark_sensitive(ds, 'f*')
    all_meta = dict(ds.repo.get_metadata('.'))
    target_rec = {'distribution-restrictions': ['sensitive']}
    # g2 since the same content
    assert not all_meta.pop('g1', None)  # nothing or empty record
    assert all_meta == {'f1': target_rec, 'f2': target_rec, 'g2': target_rec}



#! /usr/bin/env python 

import sys
import re
import ctypes


def bids_id_(sid):
    parsed_id = re.compile(r"^(?:sub-|)(.+)$").search(sid).group(1)
    return str(ctypes.c_size_t(hash(parsed_id)).value)


def main():
    sid = sys.argv[1]
    return bids_id_(sid)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import logging
import os
import sys
from argparse import ArgumentParser

from .. import __version__
from ..main import workflow

lgr = logging.getLogger(__name__)


def main(argv=None):
    parser = get_parser()
    args = parser.parse_args(argv)
    # exit if nothing to be done
    if not args.files and not args.dicom_dir_template and not args.command:
        lgr.warning("Nothing to be done - displaying usage help")
        parser.print_help()
        sys.exit(1)

    kwargs = vars(args)
    workflow(**kwargs)


def get_parser():
    docstr = ("""Example:
             heudiconv -d 'rawdata/{subject}' -o . -f heuristic.py -s s1 s2 s3""")
    parser = ArgumentParser(description=docstr)
    parser.add_argument('--version', action='version', version=__version__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-d', '--dicom_dir_template',
        dest='dicom_dir_template',
        help='Location of dicomdir that can be indexed with subject id '
             '{subject} and session {session}. Tarballs (can be compressed) '
             'are supported in addition to directory. All matching tarballs '
             'for a subject are extracted and their content processed in a '
             'single pass. If multiple tarballs are found, each is assumed to '
             'be a separate session and the --ses argument is ignored. Note '
             'that you might need to surround the value with quotes to avoid '
             '{...} being considered by shell')
    group.add_argument(
        '--files',
        nargs='*',
        help='Files (tarballs, dicoms) or directories containing files to '
             'process. Cannot be provided if using --dicom_dir_template.')
    parser.add_argument(
        '-s', '--subjects',
        dest='subjs',
        type=str,
        nargs='*',
        help='List of subjects - required for dicom template. If not '
             'provided, DICOMS would first be "sorted" and subject IDs '
             'deduced by the heuristic.')
    parser.add_argument(
        '-c', '--converter',
        choices=('dcm2niix', 'none'),
        default='dcm2niix',
        help='Tool to use for DICOM conversion. Setting to "none" disables '
             'the actual conversion step -- useful for testing heuristics.')
    parser.add_argument(
        '-o', '--outdir',
        default=os.getcwd(),
        help='Output directory for conversion setup (for further '
             'customization and future reference. This directory will refer '
             'to non-anonymized subject IDs.')
    parser.add_argument(
        '-l', '--locator',
        default=None,
        help='Study path under outdir. If provided, it overloads the value '
             'provided by the heuristic. If --datalad is enabled, every '
             'directory within locator becomes a super-dataset thus '
             'establishing a hierarchy. Setting to "unknown" will skip that '
             'dataset.')
    parser.add_argument(
        '-a', '--conv-outdir',
        default=None,
        help='Output directory for converted files. By default this is '
             'identical to --outdir. This option is most useful in '
             'combination with --anon-cmd.')
    parser.add_argument(
        '--anon-cmd',
        default=None,
        help='Command to run to convert subject IDs used for DICOMs to '
             'anonymized IDs. Such command must take a single argument and '
             'return a single anonymized ID. Also see --conv-outdir.')
    parser.add_argument(
        '-f', '--heuristic',
        dest='heuristic',
        help='Name of a known heuristic or path to the Python script '
             'containing heuristic.')
    parser.add_argument(
        '-p', '--with-prov',
        action='store_true',
        help='Store additional provenance information. Requires python-rdflib.')
    parser.add_argument(
        '-ss', '--ses',
        dest='session',
        default=None,
        help='Session for longitudinal study_sessions. Default is None.')
    parser.add_argument(
        '-b', '--bids',
        nargs='*',
        metavar=('BIDSOPTION1', 'BIDSOPTION2'),
        choices=['notop'],
        dest='bids_options',
        help='Flag for output into BIDS structure. Can also take BIDS-'
             'specific options, e.g., --bids notop. The only currently '
             'supported options is "notop", which skips creation of '
             'top-level BIDS files. This is useful when running in batch '
             'mode to prevent possible race conditions.')
    parser.add_argument(
        '--overwrite',
        action='store_true',
        default=False,
        help='Overwrite existing converted files.')
    parser.add_argument(
        '--datalad',
        action='store_true',
        help='Store the entire collection as DataLad dataset(s). Small files '
             'will be committed directly to git, while large to annex. New '
             'version (6) of annex repositories will be used in a "thin" '
             'mode so it would look to mortals as just any other regular '
             'directory (i.e. no symlinks to under .git/annex). For now just '
             'for BIDS mode.')
    parser.add_argument(
        '--dbg',
        action='store_true',
        dest='debug',
        help='Do not catch exceptions and show exception traceback.')
    parser.add_argument(
        '--command',
        choices=(
            'heuristics', 'heuristic-info',
            'ls', 'populate-templates',
            'sanitize-jsons', 'treat-jsons',
        ),
        help='Custom action to be performed on provided files instead of '
             'regular operation.')
    parser.add_argument(
        '-g', '--grouping',
        default='studyUID',
        choices=('studyUID', 'accession_number', 'all', 'custom'),
        help='How to group dicoms (default: by studyUID).')
    parser.add_argument(
        '--minmeta',
        action='store_true',
        help='Exclude dcmstack meta information in sidecar jsons.')
    parser.add_argument(
        '--random-seed',
        type=int,
        default=None,
        help='Random seed to initialize RNG.')
    parser.add_argument(
        '--dcmconfig',
        default=None,
        help='JSON file for additional dcm2niix configuration.')
    submission = parser.add_argument_group('Conversion submission options')
    submission.add_argument(
        '-q', '--queue',
        choices=("SLURM", None),
        default=None,
        help='Batch system to submit jobs in parallel.')
    submission.add_argument(
        '--queue-args',
        dest='queue_args',
        default=None,
        help='Additional queue arguments passed as a single string of '
             'space-separated Argument=Value pairs.')
    return parser


if __name__ == "__main__":
    main()



import os
import pytest
import sys
import time

from mock import patch
from os.path import join as opj
from os.path import dirname
from six.moves import StringIO
from glob import glob

from heudiconv.dicoms import compress_dicoms
from heudiconv.utils import TempDirs, file_md5sum

tests_datadir = opj(dirname(__file__), 'data')


def test_reproducibility(tmpdir):
    prefix = str(tmpdir.join("precious"))
    args = [glob(opj(tests_datadir, '01-fmap_acq-3mm', '*')),
            prefix,
            TempDirs(),
            True]
    tarball = compress_dicoms(*args)
    md5 = file_md5sum(tarball)
    assert tarball
    # must not override, ensure overwrite is set to False
    args[-1] = False
    assert compress_dicoms(*args) is None
    # reset this
    args[-1] = True

    os.unlink(tarball)

    time.sleep(1.1)  # need to guarantee change of time
    tarball_ = compress_dicoms(*args)
    md5_ = file_md5sum(tarball_)
    assert tarball == tarball_
    assert md5 == md5_



import os
import sys
import subprocess

from heudiconv.cli.run import main as runner
from heudiconv.queue import clean_args, which
from .utils import TESTS_DATA_PATH
import pytest

@pytest.mark.skipif(bool(which("sbatch")), reason="skip a real slurm call")
@pytest.mark.parametrize(
    'invocation', [
        "--files %s/01-fmap_acq-3mm" % TESTS_DATA_PATH,    # our new way with automated groupping
        "-d %s/{subject}/* -s 01-fmap_acq-3mm" % TESTS_DATA_PATH # "old" way specifying subject
    ])
def test_queue_no_slurm(tmpdir, invocation):
    tmpdir.chdir()
    hargs = invocation.split(" ")
    hargs.extend(["-f", "reproin", "-b", "--minmeta", "--queue", "SLURM"])

    # simulate command-line call
    _sys_args = sys.argv
    sys.argv = ['heudiconv'] + hargs

    try:
        with pytest.raises(OSError):  # SLURM should not be installed
            runner(hargs)
        # should have generated a slurm submission script
        slurm_cmd_file = (tmpdir / 'heudiconv-SLURM.sh').strpath
        assert slurm_cmd_file
        # check contents and ensure args match
        with open(slurm_cmd_file) as fp:
            lines = fp.readlines()
        assert lines[0] == "#!/bin/bash\n"
        cmd = lines[1]

        # check that all flags we gave still being called
        for arg in hargs:
            # except --queue <queue>
            if arg in ['--queue', 'SLURM']:
                assert arg not in cmd
            else:
                assert arg in cmd
    finally:
        # revert before breaking something
        sys.argv = _sys_args

def test_argument_filtering(tmpdir):
    cmd_files = [
        'heudiconv',
        '--files',
        '/fake/path/to/files',
        '/another/fake/path',
        '-f',
        'convertall',
        '-q',
        'SLURM',
        '--queue-args',
        '--cpus-per-task=4 --contiguous --time=10'
    ]
    filtered = [
        'heudiconv',
        '--files',
        '/another/fake/path',
        '-f',
        'convertall',
    ]
    assert clean_args(cmd_files, 'files', 1) == filtered

    cmd_subjects = [
        'heudiconv',
        '-d',
        '/some/{subject}/path',
        '--queue',
        'SLURM',
        '--subjects',
        'sub1',
        'sub2',
        'sub3',
        'sub4',
        '-f',
        'convertall'
    ]
    filtered = [
        'heudiconv',
        '-d',
        '/some/{subject}/path',
        '--subjects',
        'sub3',
        '-f',
        'convertall'
    ]
    assert clean_args(cmd_subjects, 'subjects', 2) == filtered



from collections import namedtuple
import os
import os.path as op
import pytest
from mock import patch
try:
    from tinydb import TinyDB, Query
except ImportError:
    pytest.importorskip("tinydb")
from subprocess import CalledProcessError

try:
    from heudiconv.cli.monitor import (monitor, process, run_heudiconv,
                                       MASK_NEWDIR)

    Header = namedtuple('header', ['wd', 'mask', 'cookie', 'len'])
    header = Header(5, MASK_NEWDIR, 5, 5)
    watch_path = b'WATCHME'
    filename = b'FILE'
    type_names = b'TYPE'

    path2 = watch_path + b'/' + filename + b'/subpath'

    my_events = [(header, type_names, watch_path, filename),
                 (header, type_names, path2, b'')]
except AttributeError:
    # Import of inotify fails on mac os x with error
    # lsym(0x11fbeb780, inotify_init): symbol not found
    # because inotify doesn't seem to exist on Mac OS X
    my_events = []
    pytestmark = pytest.mark.skip(reason='Unable to import inotify')


class MockInotifyTree(object):
    def __init__(self, events):
        self.events = iter(events)
    def event_gen(self):
        for e in self.events:
            yield e
    def __call__(self, topdir):
        return self


class MockTime(object):
    def __init__(self, time):
        self.time = time
    def __call__(self):
        return self.time



@pytest.mark.skip(reason="TODO")
@patch('inotify.adapters.InotifyTree', MockInotifyTree(my_events))
@patch('time.time', MockTime(42))
def test_monitor(capsys):
    monitor(watch_path.decode(), check_ptrn='')
    out, err = capsys.readouterr()
    desired_output = '{0}/{1} {2}\n'.format(watch_path.decode(),
                                            filename.decode(),
                                            42)
    desired_output += 'Updating {0}/{1}: {2}\n'.format(watch_path.decode(),
                                                       filename.decode(),
                                                       42)
    assert out == desired_output


@pytest.mark.skip(reason="TODO")
@patch('time.time', MockTime(42))
@pytest.mark.parametrize("side_effect,success", [
    (None, 1),
    (CalledProcessError('mycmd', 1), 0)
])
def test_process(tmpdir, capsys, side_effect, success):
    db_fn = tmpdir.join('database.json')
    log_dir = tmpdir.mkdir('log')
    db = TinyDB(db_fn.strpath)
    process_me = '/my/path/A12345'
    accession_number = op.basename(process_me)
    paths2process = {process_me: 42}
    with patch('subprocess.Popen') as mocked_popen:
        stdout = b"INFO: PROCESSING STARTS: {'just': 'a test'}"
        mocked_popen_instance = mocked_popen.return_value
        mocked_popen_instance.side_effect = side_effect
        mocked_popen_instance.communicate.return_value = (stdout, )
        # set return value for wait
        mocked_popen_instance.wait.return_value = 1 - success
        # mock also communicate to get the supposed stdout
        process(paths2process, db, wait=-30, logdir=log_dir.strpath)
        out, err = capsys.readouterr()
        log_fn = log_dir.join(accession_number + '.log')

        mocked_popen.assert_called_once()
        assert log_fn.check()
        assert log_fn.read() == stdout.decode('utf-8')
        assert db_fn.check()
        # dictionary should be empty
        assert not paths2process
        assert out == 'Time to process {0}\n'.format(process_me)

        # check what we have in the database
        Path = Query()
        query = db.get(Path.input_path == process_me)
        assert len(db) == 1
        assert query
        assert query['success'] == success
        assert query['accession_number'] == op.basename(process_me)
        assert query['just'] == 'a test'

@pytest.mark.skip(reason="TODO")
def test_run_heudiconv():
    # echo should succeed always
    mydict = {'key1': 'value1', 'key2': 'value2', 'success': 1}
    cmd = "echo INFO: PROCESSING STARTS: {0}".format(str(mydict))
    stdout, info_dict = run_heudiconv(cmd)
    assert info_dict == mydict
    assert "echo " + stdout.strip() == cmd



import subprocess
import sys
import os
import logging

from .utils import which

lgr = logging.getLogger(__name__)

def queue_conversion(queue, iterarg, iterables, queue_args=None):
    """
    Write out conversion arguments to file and submit to a job scheduler.
    Parses `sys.argv` for heudiconv arguments.

    Parameters
    ----------
    queue: string
        Batch scheduler to use
    iterarg: str
        Multi-argument to index (`subjects` OR `files`)
    iterables: int
        Number of `iterarg` arguments
    queue_args: string (optional)
        Additional queue arguments for job submission

    """

    SUPPORTED_QUEUES = {'SLURM': 'sbatch'}
    if queue not in SUPPORTED_QUEUES:
        raise NotImplementedError("Queuing with %s is not supported", queue)

    for i in range(iterables):
        args = clean_args(sys.argv[1:], iterarg, i)
        # make arguments executable
        heudiconv_exec = which("heudiconv") or "heudiconv"
        args.insert(0, heudiconv_exec)
        convertcmd = " ".join(args)

        # will overwrite across subjects
        queue_file = os.path.abspath('heudiconv-%s.sh' % queue)
        with open(queue_file, 'wt') as fp:
            fp.write("#!/bin/bash\n")
            if queue_args:
                for qarg in queue_args.split():
                    fp.write("#SBATCH %s\n" % qarg)
            fp.write(convertcmd + "\n")

        cmd = [SUPPORTED_QUEUES[queue], queue_file]
        proc = subprocess.call(cmd)
    lgr.info("Submitted %d jobs", iterables)

def clean_args(hargs, iterarg, iteridx):
    """
    Filters arguments for batch submission.

    Parameters
    ----------
    hargs: list
        Command-line arguments
    iterarg: str
        Multi-argument to index (`subjects` OR `files`)
    iteridx: int
        `iterarg` index to submit

    Returns
    -------
    cmdargs : list
        Filtered arguments for batch submission

    Example
    --------
    >>> from heudiconv.queue import clean_args
    >>> cmd = ['heudiconv', '-d', '/some/{subject}/path',
    ...                     '-q', 'SLURM',
    ...                     '-s', 'sub-1', 'sub-2', 'sub-3', 'sub-4']
    >>> clean_args(cmd, 'subjects', 0)
    ['heudiconv', '-d', '/some/{subject}/path', '-s', 'sub-1']
    """

    if iterarg == "subjects":
        iterarg = ['-s', '--subjects']
    elif iterarg == "files":
        iterarg = ['--files']
    else:
        raise ValueError("Cannot index %s" % iterarg)

    # remove these or cause an infinite loop
    queue_args = ['-q', '--queue', '--queue-args']

    # control variables for multi-argument parsing
    is_iterarg = False
    itercount = 0

    indicies = []
    cmdargs = hargs[:]

    for i, arg in enumerate(hargs):
        if arg.startswith('-') and is_iterarg:
            # moving on to another argument
            is_iterarg = False
        if is_iterarg:
            if iteridx != itercount:
                indicies.append(i)
            itercount += 1
        if arg in iterarg:
            is_iterarg = True
        if arg in queue_args:
            indicies.extend([i, i+1])

    for j in sorted(indicies, reverse=True):
        del cmdargs[j]
    return cmdargs



import json
import os
import os.path as op

from heudiconv.utils import (
    get_known_heuristics_with_descriptions,
    get_heuristic_description,
    load_heuristic,
    json_dumps_pretty,
    load_json,
    create_tree,
    save_json,
    get_datetime,
    JSONDecodeError)

import pytest
from .utils import HEURISTICS_PATH


def test_get_known_heuristics_with_descriptions():
    d = get_known_heuristics_with_descriptions()
    assert {'reproin', 'convertall'}.issubset(d)
    # ATM we include all, not only those two
    assert len(d) > 2
    assert len(d['reproin']) > 50  # it has a good one
    assert len(d['reproin'].split(os.sep)) == 1  # but just one line


def test_get_heuristic_description():
    desc = get_heuristic_description('reproin', full=True)
    assert len(desc) > 1000
    # and we describe such details as
    assert '_ses-' in desc
    assert '_run-' in desc
    # and mention ReproNim ;)
    assert 'ReproNim' in desc


def test_load_heuristic():
    by_name = load_heuristic('reproin')
    from_file = load_heuristic(op.join(HEURISTICS_PATH, 'reproin.py'))

    assert by_name
    assert by_name.filename == from_file.filename

    with pytest.raises(ImportError):
        load_heuristic('unknownsomething')

    with pytest.raises(ImportError):
        load_heuristic(op.join(HEURISTICS_PATH, 'unknownsomething.py'))


def test_json_dumps_pretty():
    pretty = json_dumps_pretty
    assert pretty({"SeriesDescription": "Trace:Nov 13 2017 14-36-14 EST"}) \
        == '{\n  "SeriesDescription": "Trace:Nov 13 2017 14-36-14 EST"\n}'
    assert pretty({}) == "{}"
    assert pretty({"a": -1, "b": "123", "c": [1, 2, 3], "d": ["1.0", "2.0"]}) \
        == '{\n  "a": -1,\n  "b": "123",\n  "c": [1, 2, 3],\n  "d": ["1.0", "2.0"]\n}'
    assert pretty({'a': ["0.3", "-1.9128906358217845e-12", "0.2"]}) \
        == '{\n  "a": ["0.3", "-1.9128906358217845e-12", "0.2"]\n}'
    # original, longer string
    tstr = 'f9a7d4be-a7d7-47d2-9de0-b21e9cd10755||' \
          'Sequence: ve11b/master r/50434d5; ' \
          'Mar  3 2017 10:46:13 by eja'
    # just the date which reveals the issue
    # tstr = 'Mar  3 2017 10:46:13 by eja'
    assert pretty({'WipMemBlock': tstr}) == '{\n  "WipMemBlock": "%s"\n}' % tstr


def test_load_json(tmpdir, caplog):
    # test invalid json
    ifname = 'invalid.json'
    invalid_json_file = str(tmpdir / ifname)
    create_tree(str(tmpdir), {ifname: u"I'm Jason Bourne"})

    with pytest.raises(JSONDecodeError):
        load_json(str(invalid_json_file))

    assert ifname in caplog.text

    # test valid json
    vcontent = {"secret": "spy"}
    vfname = "valid.json"
    valid_json_file = str(tmpdir / vfname)
    save_json(valid_json_file, vcontent)
    
    assert load_json(valid_json_file) == vcontent


def test_get_datetime():
    """
    Test utils.get_datetime()
    """
    assert get_datetime('20200512', '162130') == '2020-05-12T16:21:30'
    assert get_datetime('20200512', '162130.5') == '2020-05-12T16:21:30.500000'
    assert get_datetime('20200512', '162130.5', microseconds=False) == '2020-05-12T16:21:30'



#
# Tests for reproin.py
#
from collections import OrderedDict
from mock import patch
import re

from . import reproin
from .reproin import (
    filter_files,
    fix_canceled_runs,
    fix_dbic_protocol,
    fixup_subjectid,
    get_dups_marked,
    md5sum,
    parse_series_spec,
    sanitize_str,
)


def test_get_dups_marked():
    no_dups = {('some',): [1]}
    assert get_dups_marked(no_dups) == no_dups

    info = OrderedDict(
        [
            (('bu', 'du'), [1, 2]),
            (('smth',), [3]),
            (('smth2',), ['a', 'b', 'c'])
         ]
    )

    assert get_dups_marked(info) == get_dups_marked(info, True) == \
        {
            ('bu__dup-01', 'du'): [1],
            ('bu', 'du'): [2],
            ('smth',): [3],
            ('smth2__dup-01',): ['a'],
            ('smth2__dup-02',): ['b'],
            ('smth2',): ['c']
        }

    assert get_dups_marked(info, per_series=False) == \
        {
            ('bu__dup-01', 'du'): [1],
            ('bu', 'du'): [2],
            ('smth',): [3],
            ('smth2__dup-02',): ['a'],
            ('smth2__dup-03',): ['b'],
            ('smth2',): ['c']
        }



def test_filter_files():
    # Filtering is currently disabled -- any sequence directory is Ok
    assert(filter_files('/home/mvdoc/dbic/09-run_func_meh/0123432432.dcm'))
    assert(filter_files('/home/mvdoc/dbic/run_func_meh/012343143.dcm'))


def test_md5sum():
    assert md5sum('cryptonomicon') == '1cd52edfa41af887e14ae71d1db96ad1'
    assert md5sum('mysecretmessage') == '07989808231a0c6f522f9d8e34695794'


def test_fix_canceled_runs():
    from collections import namedtuple
    FakeSeqInfo = namedtuple('FakeSeqInfo',
                             ['accession_number', 'series_id',
                              'protocol_name', 'series_description'])

    seqinfo = []
    runname = 'func_run+'
    for i in range(1, 6):
        seqinfo.append(
            FakeSeqInfo('accession1',
                        '{0:02d}-'.format(i) + runname,
                        runname, runname)
        )

    fake_accession2run = {
        'accession1': ['^01-', '^03-']
    }

    with patch.object(reproin, 'fix_accession2run', fake_accession2run):
        seqinfo_ = fix_canceled_runs(seqinfo)

    for i, s in enumerate(seqinfo_, 1):
        output = runname
        if i == 1 or i == 3:
            output = 'cancelme_' + output
        for key in ['series_description', 'protocol_name']:
            value = getattr(s, key)
            assert(value == output)
        # check we didn't touch series_id
        assert(s.series_id == '{0:02d}-'.format(i) + runname)


def test_fix_dbic_protocol():
    from collections import namedtuple
    FakeSeqInfo = namedtuple('FakeSeqInfo',
                             ['accession_number', 'study_description',
                              'field1', 'field2'])
    accession_number = 'A003'
    seq1 = FakeSeqInfo(accession_number,
                       'mystudy',
                       '02-anat-scout_run+_MPR_sag',
                       '11-func_run-life2_acq-2mm692')
    seq2 = FakeSeqInfo(accession_number,
                       'mystudy',
                       'nochangeplease',
                       'nochangeeither')

    seqinfos = [seq1, seq2]
    protocols2fix = {
        md5sum('mystudy'):
            [('scout_run\+', 'THESCOUT-runX'),
             ('run-life[0-9]', 'run+_task-life')],
        re.compile('^my.*'):
            [('THESCOUT-runX', 'THESCOUT')],
        # rely on 'catch-all' to fix up above scout
        '': [('THESCOUT', 'scout')]
    }

    with patch.object(reproin, 'protocols2fix', protocols2fix), \
            patch.object(reproin, 'series_spec_fields', ['field1']):
        seqinfos_ = fix_dbic_protocol(seqinfos)
    assert(seqinfos[1] == seqinfos_[1])
    # field2 shouldn't have changed since I didn't pass it
    assert(seqinfos_[0] == FakeSeqInfo(accession_number,
                                       'mystudy',
                                       '02-anat-scout_MPR_sag',
                                       seq1.field2))

    # change also field2 please
    with patch.object(reproin, 'protocols2fix', protocols2fix), \
            patch.object(reproin, 'series_spec_fields', ['field1', 'field2']):
        seqinfos_ = fix_dbic_protocol(seqinfos)
    assert(seqinfos[1] == seqinfos_[1])
    # now everything should have changed
    assert(seqinfos_[0] == FakeSeqInfo(accession_number,
                                       'mystudy',
                                       '02-anat-scout_MPR_sag',
                                       '11-func_run+_task-life_acq-2mm692'))


def test_sanitize_str():
    assert sanitize_str('super@duper.faster') == 'superduperfaster'
    assert sanitize_str('perfect') == 'perfect'
    assert sanitize_str('never:use:colon:!') == 'neverusecolon'


def test_fixupsubjectid():
    assert fixup_subjectid("abra") == "abra"
    assert fixup_subjectid("sub") == "sub"
    assert fixup_subjectid("sid") == "sid"
    assert fixup_subjectid("sid000030") == "sid000030"
    assert fixup_subjectid("sid0000030") == "sid000030"
    assert fixup_subjectid("sid00030") == "sid000030"
    assert fixup_subjectid("sid30") == "sid000030"
    assert fixup_subjectid("SID30") == "sid000030"


def test_parse_series_spec():
    pdpn = parse_series_spec

    assert pdpn("nondbic_func-bold") == {}
    assert pdpn("cancelme_func-bold") == {}

    assert pdpn("bids_func-bold") == \
           pdpn("func-bold") == \
           {'seqtype': 'func', 'seqtype_label': 'bold'}

    # pdpn("bids_func_ses+_task-boo_run+") == \
    # order and PREFIX: should not matter, as well as trailing spaces
    assert \
        pdpn(" PREFIX:bids_func_ses+_task-boo_run+  ") == \
        pdpn("PREFIX:bids_func_ses+_task-boo_run+") == \
        pdpn("WIP func_ses+_task-boo_run+") == \
        pdpn("bids_func_ses+_run+_task-boo") == \
           {
               'seqtype': 'func',
               # 'seqtype_label': 'bold',
               'session': '+',
               'run': '+',
               'task': 'boo',
            }

    # TODO: fix for that
    assert pdpn("bids_func-pace_ses-1_task-boo_acq-bu_bids-please_run-2__therest") == \
           pdpn("bids_func-pace_ses-1_run-2_task-boo_acq-bu_bids-please__therest") == \
           pdpn("func-pace_ses-1_task-boo_acq-bu_bids-please_run-2") == \
           {
               'seqtype': 'func', 'seqtype_label': 'pace',
               'session': '1',
               'run': '2',
               'task': 'boo',
               'acq': 'bu',
               'bids': 'bids-please'
           }

    assert pdpn("bids_anat-scout_ses+") == \
           {
               'seqtype': 'anat',
               'seqtype_label': 'scout',
               'session': '+',
           }

    assert pdpn("anat_T1w_acq-MPRAGE_run+") == \
           {
                'seqtype': 'anat',
                'run': '+',
                'acq': 'MPRAGE',
                'seqtype_label': 'T1w'
           }

    # Check for currently used {date}, which should also should get adjusted
    # from (date) since Philips does not allow for {}
    assert pdpn("func_ses-{date}") == \
           pdpn("func_ses-(date)") == \
           {'seqtype': 'func', 'session': '{date}'}

    assert pdpn("fmap_dir-AP_ses-01") == \
           {'seqtype': 'fmap', 'session': '01', 'dir': 'AP'}


"""Handle BIDS specific operations"""

import hashlib
import os
import os.path as op
import logging
import re
from collections import OrderedDict
from datetime import datetime
import csv
from random import sample
from glob import glob

from .external.pydicom import dcm

from .parser import find_files
from .utils import (
    load_json,
    save_json,
    create_file_if_missing,
    json_dumps_pretty,
    set_readonly,
    is_readonly,
    get_datetime,
)

lgr = logging.getLogger(__name__)

# Fields to be populated in _scans files. Order matters
SCANS_FILE_FIELDS = OrderedDict([
    ("filename", OrderedDict([
        ("Description", "Name of the nifti file")])),
    ("acq_time", OrderedDict([
        ("LongName", "Acquisition time"),
        ("Description", "Acquisition time of the particular scan")])),
    ("operator", OrderedDict([
        ("Description", "Name of the operator")])),
    ("randstr", OrderedDict([
        ("LongName", "Random string"),
        ("Description", "md5 hash of UIDs")])),
])

class BIDSError(Exception):
    pass


BIDS_VERSION = "1.4.1"


def populate_bids_templates(path, defaults={}):
    """Premake BIDS text files with templates"""

    lgr.info("Populating template files under %s", path)
    descriptor = op.join(path, 'dataset_description.json')
    if not op.lexists(descriptor):
        save_json(descriptor,
              OrderedDict([
                  ('Name', "TODO: name of the dataset"),
                  ('BIDSVersion', BIDS_VERSION),
                  ('License', defaults.get('License',
                        "TODO: choose a license, e.g. PDDL "
                        "(http://opendatacommons.org/licenses/pddl/)")),
                  ('Authors', defaults.get('Authors',
                        ["TODO:", "First1 Last1", "First2 Last2", "..."])),
                  ('Acknowledgements', defaults.get('Acknowledgements',
                        'TODO: whom you want to acknowledge')),
                  ('HowToAcknowledge',
                        "TODO: describe how to acknowledge -- either cite a "
                        "corresponding paper, or just in acknowledgement "
                        "section"),
                  ('Funding', ["TODO", "GRANT #1", "GRANT #2"]),
                  ('ReferencesAndLinks',
                        ["TODO", "List of papers or websites"]),
                  ('DatasetDOI', 'TODO: eventually a DOI for the dataset')
        ]))
    sourcedata_README = op.join(path, 'sourcedata', 'README')
    if op.exists(op.dirname(sourcedata_README)):
        create_file_if_missing(sourcedata_README,
            ("TODO: Provide description about source data, e.g. \n"
            "Directory below contains DICOMS compressed into tarballs per "
            "each sequence, replicating directory hierarchy of the BIDS dataset"
            " itself."))
    create_file_if_missing(op.join(path, 'CHANGES'),
        "0.0.1  Initial data acquired\n"
        "TODOs:\n\t- verify and possibly extend information in participants.tsv"
        " (see for example http://datasets.datalad.org/?dir=/openfmri/ds000208)"
        "\n\t- fill out dataset_description.json, README, sourcedata/README"
        " (if present)\n\t- provide _events.tsv file for each _bold.nii.gz with"
        " onsets of events (see  '8.5 Task events'  of BIDS specification)")
    create_file_if_missing(op.join(path, 'README'),
        "TODO: Provide description for the dataset -- basic details about the "
        "study, possibly pointing to pre-registration (if public or embargoed)")

    populate_aggregated_jsons(path)


def populate_aggregated_jsons(path):
    """Aggregate across the entire BIDS dataset .json's into top level .json's

    Top level .json files would contain only the fields which are
    common to all subject[/session]/type/*_modality.json's.

    ATM aggregating only for *_task*_bold.json files. Only the task- and
    OPTIONAL _acq- field is retained within the aggregated filename.  The other
    BIDS _key-value pairs are "aggregated over".

    Parameters
    ----------
    path: str
      Path to the top of the BIDS dataset
    """
    # TODO: collect all task- .json files for func files to
    tasks = {}
    # way too many -- let's just collect all which are the same!
    # FIELDS_TO_TRACK = {'RepetitionTime', 'FlipAngle', 'EchoTime',
    #                    'Manufacturer', 'SliceTiming', ''}
    for fpath in find_files('.*_task-.*\_bold\.json', topdir=path,
                            exclude_vcs=True,
                            exclude="/\.(datalad|heudiconv)/"):
        #
        # According to BIDS spec I think both _task AND _acq (may be more?
        # _rec, _dir, ...?) should be retained?
        # TODO: if we are to fix it, then old ones (without _acq) should be
        # removed first
        task = re.sub('.*_(task-[^_\.]*(_acq-[^_\.]*)?)_.*', r'\1', fpath)
        json_ = load_json(fpath)
        if task not in tasks:
            tasks[task] = json_
        else:
            rec = tasks[task]
            # let's retain only those fields which have the same value
            for field in sorted(rec):
                if field not in json_ or json_[field] != rec[field]:
                    del rec[field]
        # create a stub onsets file for each one of those
        suf = '_bold.json'
        assert fpath.endswith(suf)
        # specify the name of the '_events.tsv' file:
        if '_echo-' in fpath:
            # multi-echo sequence: bids (1.1.0) specifies just one '_events.tsv'
            #   file, common for all echoes.  The name will not include _echo-.
            # TODO: RF to use re.match for better readability/robustness
            # So, find out the echo number:
            fpath_split = fpath.split('_echo-', 1)         # split fpath using '_echo-'
            fpath_split_2 = fpath_split[1].split('_', 1)   # split the second part of fpath_split using '_'
            echoNo = fpath_split_2[0]                      # get echo number
            if echoNo == '1':
                if len(fpath_split_2) != 2:
                    raise ValueError("Found no trailer after _echo-")
                # we modify fpath to exclude '_echo-' + echoNo:
                fpath = fpath_split[0] + '_' + fpath_split_2[1]
            else:
                # for echoNo greater than 1, don't create the events file, so go to
                #   the next for loop iteration:
                continue

        events_file = fpath[:-len(suf)] + '_events.tsv'
        # do not touch any existing thing, it may be precious
        if not op.lexists(events_file):
            lgr.debug("Generating %s", events_file)
            with open(events_file, 'w') as f:
                f.write(
                    "onset\tduration\ttrial_type\tresponse_time\tstim_file"
                    "\tTODO -- fill in rows and add more tab-separated "
                    "columns if desired")
    # extract tasks files stubs
    for task_acq, fields in tasks.items():
        task_file = op.join(path, task_acq + '_bold.json')
        # Since we are pulling all unique fields we have to possibly
        # rewrite this file to guarantee consistency.
        # See https://github.com/nipy/heudiconv/issues/277 for a usecase/bug
        # when we didn't touch existing one.
        # But the fields we enter (TaskName and CogAtlasID) might need need
        # to be populated from the file if it already exists
        placeholders = {
            "TaskName": ("TODO: full task name for %s" %
                         task_acq.split('_')[0].split('-')[1]),
            "CogAtlasID": "http://www.cognitiveatlas.org/task/id/TODO",
        }
        if op.lexists(task_file):
            j = load_json(task_file)
            # Retain possibly modified placeholder fields
            for f in placeholders:
                if f in j:
                    placeholders[f] = j[f]
            act = "Regenerating"
        else:
            act = "Generating"
        lgr.debug("%s %s", act, task_file)
        fields.update(placeholders)
        save_json(task_file, fields, sort_keys=True, pretty=True)


def tuneup_bids_json_files(json_files):
    """Given a list of BIDS .json files, e.g. """
    if not json_files:
        return
    # Harmonize generic .json formatting
    for jsonfile in json_files:
        json_ = load_json(jsonfile)
        # sanitize!
        for f1 in ['Acquisition', 'Study', 'Series']:
            for f2 in ['DateTime', 'Date']:
                json_.pop(f1 + f2, None)
        # TODO:  should actually be placed into series file which must
        #        go under annex (not under git) and marked as sensitive
        # MG - Might want to replace with flag for data sensitivity
        # related - https://github.com/nipy/heudiconv/issues/92
        if 'Date' in str(json_):
            # Let's hope no word 'Date' comes within a study name or smth like
            # that
            raise ValueError("There must be no dates in .json sidecar")
        save_json(jsonfile, json_)

    # Load the beast
    seqtype = op.basename(op.dirname(jsonfile))

    # MG - want to expand this for other _epi
    # possibly add IntendedFor automatically as well?
    if seqtype == 'fmap':
        json_basename = '_'.join(jsonfile.split('_')[:-1])
        # if we got by now all needed .json files -- we can fix them up
        # unfortunately order of "items" is not guaranteed atm
        json_phasediffname = json_basename + '_phasediff.json'
        json_mag = json_basename + '_magnitude*.json'
        if op.exists(json_phasediffname) and len(glob(json_mag)) >= 1:
            json_ = load_json(json_phasediffname)
            # TODO: we might want to reorder them since ATM
            # the one for shorter TE is the 2nd one!
            # For now just save truthfully by loading magnitude files
            lgr.debug("Placing EchoTime fields into phasediff file")
            for i in 1, 2:
                try:
                    json_['EchoTime%d' % i] = (load_json(json_basename +
                                          '_magnitude%d.json' % i)['EchoTime'])
                except IOError as exc:
                    lgr.error("Failed to open magnitude file: %s", exc)
            # might have been made R/O already, but if not -- it will be set
            # only later in the pipeline, so we must not make it read-only yet
            was_readonly = is_readonly(json_phasediffname)
            if was_readonly:
                set_readonly(json_phasediffname, False)
            save_json(json_phasediffname, json_)
            if was_readonly:
                set_readonly(json_phasediffname)


def add_participant_record(studydir, subject, age, sex):
    participants_tsv = op.join(studydir, 'participants.tsv')
    participant_id = 'sub-%s' % subject

    if not create_file_if_missing(participants_tsv,
           '\t'.join(['participant_id', 'age', 'sex', 'group']) + '\n'):
        # check if may be subject record already exists
        with open(participants_tsv) as f:
            f.readline()
            known_subjects = {l.split('\t')[0] for l in f.readlines()}
        if participant_id in known_subjects:
            return
    else:
        # Populate particpants.json (an optional file to describe column names in
        # participant.tsv). This auto generation will make BIDS-validator happy.
        participants_json = op.join(studydir, 'participants.json')
        if not op.lexists(participants_json):
            save_json(participants_json,
                OrderedDict([
                    ("participant_id", OrderedDict([
                        ("Description", "Participant identifier")])),
                    ("age", OrderedDict([
                        ("Description", "Age in years (TODO - verify) as in the initial"
                            " session, might not be correct for other sessions")])),
                    ("sex", OrderedDict([
                        ("Description", "self-rated by participant, M for male/F for "
                            "female (TODO: verify)")])),
                    ("group", OrderedDict([
                        ("Description", "(TODO: adjust - by default everyone is in "
                            "control group)")])),
                ]),
                sort_keys=False)
    # Add a new participant
    with open(participants_tsv, 'a') as f:
        f.write(
          '\t'.join(map(str, [participant_id,
                              age.lstrip('0').rstrip('Y') if age else 'N/A',
                              sex,
                              'control'])) + '\n')


def find_subj_ses(f_name):
    """Given a path to the bids formatted filename parse out subject/session"""
    # we will allow the match at either directories or within filename
    # assuming that bids layout is "correct"
    regex = re.compile('sub-(?P<subj>[a-zA-Z0-9]*)([/_]ses-(?P<ses>[a-zA-Z0-9]*))?')
    regex_res = regex.search(f_name)
    res = regex_res.groupdict() if regex_res else {}
    return res.get('subj', None), res.get('ses', None)


def save_scans_key(item, bids_files):
    """
    Parameters
    ----------
    item:
    bids_files: str or list

    Returns
    -------

    """
    rows = {}
    assert bids_files, "we do expect some files since it was called"
    # we will need to deduce subject and session from the bids_filename
    # and if there is a conflict, we would just blow since this function
    # should be invoked only on a result of a single item conversion as far
    # as I see it, so should have the same subject/session
    subj, ses = None, None
    for bids_file in bids_files:
        # get filenames
        f_name = '/'.join(bids_file.split('/')[-2:])
        f_name = f_name.replace('json', 'nii.gz')
        rows[f_name] = get_formatted_scans_key_row(item[-1][0])
        subj_, ses_ = find_subj_ses(f_name)
        if not subj_:
            lgr.warning(
                "Failed to detect fullfilled BIDS layout.  "
                "No scans.tsv file(s) will be produced for %s",
                ", ".join(bids_files)
            )
            return
        if subj and subj_ != subj:
            raise ValueError(
                "We found before subject %s but now deduced %s from %s"
                % (subj, subj_, f_name))
        subj = subj_
        if ses and ses_ != ses:
            raise ValueError(
                "We found before session %s but now deduced %s from %s"
                % (ses, ses_, f_name)
            )
        ses = ses_
    # where should we store it?
    output_dir = op.dirname(op.dirname(bids_file))
    # save
    ses = '_ses-%s' % ses if ses else ''
    add_rows_to_scans_keys_file(
        op.join(output_dir, 'sub-{0}{1}_scans.tsv'.format(subj, ses)), rows)


def add_rows_to_scans_keys_file(fn, newrows):
    """
    Add new rows to file fn for scans key filename and generate accompanying json
    descriptor to make BIDS validator happy.

    Parameters
    ----------
    fn: filename
    newrows: extra rows to add
        dict fn: [acquisition time, referring physician, random string]
    """
    if op.lexists(fn):
        with open(fn, 'r') as csvfile:
            reader = csv.reader(csvfile, delimiter='\t')
            existing_rows = [row for row in reader]
        # skip header
        fnames2info = {row[0]: row[1:] for row in existing_rows[1:]}

        newrows_key = newrows.keys()
        newrows_toadd = list(set(newrows_key) - set(fnames2info.keys()))
        for key_toadd in newrows_toadd:
            fnames2info[key_toadd] = newrows[key_toadd]
        # remove
        os.unlink(fn)
    else:
        fnames2info = newrows
        # Populate _scans.json (an optional file to describe column names in
        # _scans.tsv). This auto generation will make BIDS-validator happy.
        scans_json = '.'.join(fn.split('.')[:-1] + ['json'])
        if not op.lexists(scans_json):
            save_json(scans_json, SCANS_FILE_FIELDS, sort_keys=False)

    header = SCANS_FILE_FIELDS
    # prepare all the data rows
    data_rows = [[k] + v for k, v in fnames2info.items()]
    # sort by the date/filename
    try:
        data_rows_sorted = sorted(data_rows, key=lambda x: (x[1], x[0]))
    except TypeError as exc:
        lgr.warning("Sorting scans by date failed: %s", str(exc))
        data_rows_sorted = sorted(data_rows)
    # save
    with open(fn, 'a') as csvfile:
        writer = csv.writer(csvfile, delimiter='\t')
        writer.writerows([header] + data_rows_sorted)


def get_formatted_scans_key_row(dcm_fn):
    """
    Parameters
    ----------
    item

    Returns
    -------
    row: list
        [ISO acquisition time, performing physician name, random string]

    """
    dcm_data = dcm.read_file(dcm_fn, stop_before_pixels=True, force=True)
    # we need to store filenames and acquisition times
    # parse date and time of start of run acquisition and get it into isoformat
    try:
        date = dcm_data.AcquisitionDate
        time = dcm_data.AcquisitionTime
        acq_time = get_datetime(date, time)
    except (AttributeError, ValueError) as exc:
        lgr.warning("Failed to get date/time for the content: %s", str(exc))
        acq_time = ''
    # add random string
    # But let's make it reproducible by using all UIDs
    # (might change across versions?)
    randcontent = u''.join(
        [getattr(dcm_data, f) or '' for f in sorted(dir(dcm_data))
         if f.endswith('UID')]
    )
    randstr = hashlib.md5(randcontent.encode()).hexdigest()[:8]
    try:
        perfphys = dcm_data.PerformingPhysicianName
    except AttributeError:
        perfphys = ''
    row = [acq_time, perfphys, randstr]
    # empty entries should be 'n/a'
    # https://github.com/dartmouth-pbs/heudiconv/issues/32
    row = ['n/a' if not str(e) else e for e in row]
    return row


def convert_sid_bids(subject_id):
    """Strips any non-BIDS compliant characters within subject_id

    Parameters
    ----------
    subject_id : string

    Returns
    -------
    sid : string
        New subject ID
    subject_id : string
        Original subject ID
    """
    cleaner = lambda y: ''.join([x for x in y if x.isalnum()])
    sid = cleaner(subject_id)
    if not sid:
        raise ValueError(
            "Subject ID became empty after cleanup.  Please provide manually "
            "a suitable alphanumeric subject ID")
    lgr.warning('{0} contained nonalphanumeric character(s), subject '
                'ID was cleaned to be {1}'.format(subject_id, sid))
    return sid, subject_id



import os

def create_key(template, outtype=('nii.gz',), annotation_classes=None):
    if template is None or not template:
        raise ValueError('Template must be a valid format string')
    return template, outtype, annotation_classes

def infotodict(seqinfo):
    """Heuristic evaluator for determining which runs belong where
    
    allowed template fields - follow python string module: 
    
    item: index within category 
    subject: participant id 
    seqitem: run number during scanning
    subindex: sub index within group
    """
    t1w = create_key('anat/sub-{subject}_T1w')
    t2w = create_key('anat/sub-{subject}_acq-{acq}_T2w')
    flair = create_key('anat/sub-{subject}_acq-{acq}_FLAIR')
    rest = create_key('func/sub-{subject}_task-rest_acq-{acq}_run-{item:02d}_bold')

    info = {t1w: [], t2w: [], flair: [], rest: []}

    for idx, seq in enumerate(seqinfo):
        x,y,z,n_vol,protocol,dcm_dir = (seq[6], seq[7], seq[8], seq[9], seq[12], seq[3])
        # t1_mprage --> T1w
        if (z == 160) and (n_vol == 1) and ('t1_mprage' in protocol) and ('XX' not in dcm_dir):
            info[t1w] = [seq[2]]
        # t2_tse --> T2w
        if (z == 35) and (n_vol == 1) and ('t2_tse' in protocol) and ('XX' not in dcm_dir):
            info[t2w].append({'item': seq[2], 'acq': 'TSE'})
        # T2W --> T2w
        if (z == 192) and (n_vol == 1) and ('T2W' in protocol) and ('XX' not in dcm_dir):
            info[t2w].append({'item': seq[2], 'acq': 'highres'})
        # t2_tirm --> FLAIR
        if (z == 35) and (n_vol == 1) and ('t2_tirm' in protocol) and ('XX' not in dcm_dir):
            info[flair].append({'item': seq[2], 'acq': 'TIRM'}) 
        # t2_flair --> FLAIR
        if (z == 160) and (n_vol == 1) and ('t2_flair' in protocol) and ('XX' not in dcm_dir):
            info[flair].append({'item': seq[2], 'acq': 'highres'})
        # T2FLAIR --> FLAIR
        if (z == 192) and (n_vol == 1) and ('T2-FLAIR' in protocol) and ('XX' not in dcm_dir):
            info[flair].append({'item': seq[2], 'acq': 'highres'})
        # EPI (physio-matched) --> bold
        if (x == 128) and (z == 28) and (n_vol == 300) and ('EPI' in protocol) and ('XX' not in dcm_dir):
            info[rest].append({'item': seq[2], 'acq': '128px'})
        # EPI (physio-matched_NEW) --> bold
        if (x == 64) and (z == 34) and (n_vol == 300) and ('EPI' in protocol) and ('XX' not in dcm_dir):
            info[rest].append({'item': seq[2], 'acq': '64px'})
    return info



"""Testing conversion with conversion saved on datalad"""
from glob import glob
import os
import os.path as op

import pytest

from heudiconv.cli.run import main as runner
from heudiconv.external.pydicom import dcm
from heudiconv.utils import load_json
# testing utilities
from .utils import fetch_data, gen_heudiconv_args, TESTS_DATA_PATH

have_datalad = True
try:
    from datalad.support.exceptions import IncompleteResultsError
except ImportError:
    have_datalad = False


@pytest.mark.skipif(not have_datalad, reason="no datalad")
@pytest.mark.parametrize('subject', ['sub-sid000143'])
@pytest.mark.parametrize('heuristic', ['reproin.py'])
@pytest.mark.parametrize('anon_cmd', [None, 'anonymize_script.py'])
def test_conversion(tmpdir, subject, heuristic, anon_cmd):
    tmpdir.chdir()
    try:
        datadir = fetch_data(tmpdir.strpath,
                             "dbic/QA",  # path from datalad database root
                             getpath=op.join('sourcedata', subject))
    except IncompleteResultsError as exc:
        pytest.skip("Failed to fetch test data: %s" % str(exc))
    outdir = tmpdir.mkdir('out').strpath

    args = gen_heudiconv_args(
        datadir, outdir, subject, heuristic, anon_cmd,
        template=op.join('sourcedata/{subject}/*/*/*.tgz')
    )
    runner(args)  # run conversion

    # verify functionals were converted
    assert (
        glob('{}/{}/func/*'.format(outdir, subject)) ==
        glob('{}/{}/func/*'.format(datadir, subject))
    )

    # compare some json metadata
    json_ = '{}/task-rest_acq-24mm64sl1000tr32te600dyn_bold.json'.format
    orig, conv = (load_json(json_(datadir)),
                  load_json(json_(outdir)))
    keys = ['EchoTime', 'MagneticFieldStrength', 'Manufacturer', 'SliceTiming']
    for key in keys:
        assert orig[key] == conv[key]


@pytest.mark.skipif(not have_datalad, reason="no datalad")
def test_multiecho(tmpdir, subject='MEEPI', heuristic='bids_ME.py'):
    tmpdir.chdir()
    try:
        datadir = fetch_data(tmpdir.strpath, "dicoms/velasco/MEEPI")
    except IncompleteResultsError as exc:
        pytest.skip("Failed to fetch test data: %s" % str(exc))

    outdir = tmpdir.mkdir('out').strpath
    args = gen_heudiconv_args(datadir, outdir, subject, heuristic)
    runner(args)  # run conversion

    # check if we have echo functionals
    echoes = glob(op.join('out', 'sub-' + subject, 'func', '*echo*nii.gz'))
    assert len(echoes) == 3

    # check EchoTime of each functional
    # ET1 < ET2 < ET3
    prev_echo = 0
    for echo in sorted(echoes):
        _json = echo.replace('.nii.gz', '.json')
        assert _json
        echotime = load_json(_json).get('EchoTime', None)
        assert echotime > prev_echo
        prev_echo = echotime

    events = glob(op.join('out', 'sub-' + subject, 'func', '*events.tsv'))
    for event in events:
        assert 'echo-' not in event


@pytest.mark.parametrize('subject', ['merged'])
def test_grouping(tmpdir, subject):
    dicoms = [
        op.join(TESTS_DATA_PATH, fl) for fl in ['axasc35.dcm', 'phantom.dcm']
    ]
    # ensure DICOMs are different studies
    studyuids = {
        dcm.read_file(fl, stop_before_pixels=True).StudyInstanceUID for fl
        in dicoms
    }
    assert len(studyuids) == len(dicoms)
    # symlink to common location
    outdir = tmpdir.mkdir('out')
    datadir = tmpdir.mkdir(subject)
    for fl in dicoms:
        os.symlink(fl, (datadir / op.basename(fl)).strpath)

    template = op.join("{subject}/*.dcm")
    hargs = gen_heudiconv_args(
        tmpdir.strpath,
        outdir.strpath,
        subject,
        'convertall.py',
        template=template
    )

    with pytest.raises(AssertionError):
        runner(hargs)

    # group all found DICOMs under subject, despite conflicts
    hargs += ["-g", "all"]
    runner(hargs)
    assert len([fl for fl in outdir.visit(fil='run0*')]) == 4
    tsv = (outdir / 'participants.tsv')
    assert tsv.check()
    lines = tsv.open().readlines()
    assert len(lines) == 2
    assert lines[1].split('\t')[0] == 'sub-{}'.format(subject)



"""Compatibility layer for dcmstack"""

from __future__ import absolute_import

from .pydicom import dcm  # to assure that we have it one way or another

try:
    import dcmstack as ds
except ImportError as e:
    from .. import lgr
    # looks different between py2 and 3 so we go for very rudimentary matching
    e_str = str(e)
    # there were changes from how
    if not (
        ("No module" in e_str and "dicom" in e_str) or
        ('has been removed in pydicom version' in e_str)
    ):
        raise
    # a butt plug due to rename of dicom -> pydicom
    import sys
    if "dicom" in sys.modules:
        lgr.warning("unexpected happened -- dicom module already loaded")
        raise  # reraise since something is not right really
    lgr.warning("dcmstack without support of pydicom >= 1.0 is detected. Adding a plug")
    sys.modules['dicom'] = dcm
    # and try again
    import dcmstack as ds
